import datetime as dt
import unittest

import finam as fm


class TestInspect(unittest.TestCase):
    def test_slot(self):
        inp = fm.Input(
            name="In", static=True, time=None, grid=fm.UniformGrid((15, 10)), units="m"
        )
        _s1 = fm.tools.inspect(inp)

    def test_adapter(self):
        ada = fm.adapters.ValueToGrid(grid=fm.UniformGrid((20, 15)))
        _s1 = fm.tools.inspect(ada)

    def test_inspect_component(self):
        comp1 = fm.components.WeightedSum(inputs=["A", "B", "C"])
        comp2 = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None),
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=1),
        )

        comp1.initialize()
        comp2.initialize()

        _s1 = fm.tools.inspect(comp1)
        _s2 = fm.tools.inspect(comp2)


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timedelta

import numpy as np
from numpy.testing import assert_allclose

import finam as fm


class TestParametricGrid(unittest.TestCase):
    def test_parametric_1d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20,)),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x: x,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 19))
        assert_allclose(data[0].magnitude, in_info.grid.data_axes[0])

    def test_parametric_2d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15)),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x, y: x,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 19, 14))
        self.assertEqual(data[0, 0, 0], 0.5 * fm.UNITS.Unit("m"))
        self.assertEqual(data[0, 0, 2], 0.5 * fm.UNITS.Unit("m"))
        self.assertEqual(data[0, 2, 2], 2.5 * fm.UNITS.Unit("m"))

    def test_parametric_3d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15, 10)),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x, y, z: x * y - z,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 19, 14, 9))

    def test_parametric_points_1d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UnstructuredPoints(np.random.random((100, 1))),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x: x,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 100))
        assert_allclose(data[0].magnitude, [p[0] for p in in_info.grid.points])

    def test_parametric_points_2d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UnstructuredPoints(np.random.random((100, 2))),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x, y: x * y,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 100))
        assert_allclose(data[0].magnitude, [p[0] * p[1] for p in in_info.grid.points])

    def test_parametric_points_3d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UnstructuredPoints(np.random.random((100, 3))),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x, y, z: x * y - z,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 100))
        assert_allclose(
            data[0].magnitude, [p[0] * p[1] - p[2] for p in in_info.grid.points]
        )

    def test_parametric_fail_nogrid(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.NoGrid(1),
            units="m",
        )

        source = fm.components.ParametricGrid(
            info=in_info,
            func=lambda t, x: x,
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Grid"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        with self.assertRaises(fm.FinamMetaDataError):
            composition.connect(time)


class TestStaticParametricGrid(unittest.TestCase):
    def test_static_parametric(self):
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15)),
            units="m",
        )

        source = fm.components.StaticParametricGrid(
            info=in_info,
            func=lambda x, y: x,
        )
        sink = fm.components.DebugPushConsumer(
            inputs={
                "Input": fm.Info(time=None, grid=in_info.grid, units=None),
            },
            log_data="INFO",
        )

        composition = fm.Composition([source, sink])

        source.outputs["Grid"] >> sink.inputs["Input"]

        composition.connect(None)
        data_1 = sink.data["Input"][0, ...]

        self.assertEqual(data_1.shape, (19, 14))

        composition.run(end_time=None)

        data_2 = sink.data["Input"][0, ...]
        assert_allclose(data_1.magnitude, data_2.magnitude)
        self.assertEqual(data_2.shape, (19, 14))


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm


class TestSimplexNoise(unittest.TestCase):
    def test_simplex_uniform(self):
        time = datetime(2000, 1, 1)
        grid = fm.UniformGrid((25, 15), origin=(10.0, 5.0))

        source = fm.components.SimplexNoise(octaves=3, frequency=0.01, seed=4)

        consumer = fm.components.DebugConsumer(
            inputs={"Noise": fm.Info(None, grid=grid, units="")},
            start=time,
            step=timedelta(days=7),
        )

        composition = fm.Composition([source, consumer])

        _ = source.outputs["Noise"] >> consumer.inputs["Noise"]

        composition.connect(time)

        composition.run(end_time=datetime(2000, 3, 1))

        print(consumer.data["Noise"])

    def test_simplex_unstructured(self):
        time = datetime(2000, 1, 1)
        grid = fm.UnstructuredPoints(np.random.random((100, 2)) * 100)

        source = fm.components.SimplexNoise(octaves=3, frequency=0.01, seed=4)

        consumer = fm.components.DebugConsumer(
            inputs={"Noise": fm.Info(None, grid=grid, units="")},
            start=time,
            step=timedelta(days=7),
        )

        composition = fm.Composition([source, consumer])

        _ = source.outputs["Noise"] >> consumer.inputs["Noise"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 3, 1))

        print(consumer.data["Noise"])


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime
from os import path
from tempfile import TemporaryDirectory

from finam import UNITS, ComponentStatus, Info, Input, NoGrid
from finam.components.readers import CsvReader


class TestCsvReader(unittest.TestCase):
    def test_read_file(self):
        import pandas

        with TemporaryDirectory() as tmp:
            start = datetime(2000, 1, 1)
            file = path.join(tmp, "test.csv")

            data = pandas.DataFrame()
            data["T"] = [
                "2000-01-01",
                "2000-01-02",
                "2000-01-03",
                "2000-01-05",
                "2000-01-09",
                "2000-01-17",
            ]
            data["X"] = [1, 2, 3, 4, 5, 6]
            data["Y"] = [7, 8, 9, 10, 11, 12]

            data.to_csv(file, sep=";", index=False)

            reader = CsvReader(
                file,
                time_column="T",
                outputs={"X": "", "Y": "meter"},
                date_format=None,
            )
            sink1 = Input("In1")
            sink2 = Input("In2")

            reader.initialize()

            self.assertEqual(len(reader.outputs), 2)

            reader.outputs["X"] >> sink1
            reader.outputs["Y"] >> sink2

            sink1.ping()
            sink2.ping()

            reader.connect(start)
            reader.connect(start)

            sink1.exchange_info(Info(None, grid=NoGrid(), units=None))
            sink2.exchange_info(Info(None, grid=NoGrid(), units=None))

            reader.outputs["X"].get_info(Info(None, grid=NoGrid(), units=None))
            reader.outputs["Y"].get_info(Info(None, grid=NoGrid(), units=None))

            reader.connect(start)
            reader.connect(start)
            reader.validate()

            self.assertEqual(reader.time, datetime(2000, 1, 1))
            self.assertEqual(
                reader.outputs["X"].get_data(datetime(2000, 1, 1), None), 1
            )
            self.assertEqual(
                reader.outputs["Y"].get_data(datetime(2000, 1, 1), None),
                7 * UNITS.meter,
            )

            self.assertEqual(sink1.info.time, datetime(2000, 1, 1))
            self.assertEqual(sink2.info.time, datetime(2000, 1, 1))

            reader.update()

            self.assertEqual(reader.time, datetime(2000, 1, 2))
            self.assertEqual(
                reader.outputs["X"].get_data(datetime(2000, 1, 2), None), 2
            )
            self.assertEqual(
                reader.outputs["Y"].get_data(datetime(2000, 1, 2), None),
                8 * UNITS.meter,
            )

            reader.update()
            self.assertEqual(reader.time, datetime(2000, 1, 3))

            reader.update()
            self.assertEqual(reader.time, datetime(2000, 1, 5))

            reader.update()
            self.assertEqual(reader.time, datetime(2000, 1, 9))

            reader.update()
            self.assertEqual(reader.time, datetime(2000, 1, 17))

            reader.finalize()
            self.assertEqual(reader.status, ComponentStatus.FINALIZED)


if __name__ == "__main__":
    unittest.main()



"""
Components that are no simulation models.
Like IO, visualization, pre- and post-processing, etc.

See also book chapter :doc:`/finam-book/usage/known_modules` for a list of
other components that are not included in the core package.

Modules
=======

.. autosummary::
   :toctree: generated

    CallbackComponent
    CallbackGenerator
    CsvReader
    CsvWriter
    DebugConsumer
    DebugPushConsumer
    ParametricGrid
    ScheduleLogger
    SimplexNoise
    StaticCallbackGenerator
    StaticParametricGrid
    StaticSimplexNoise
    TimeTrigger
    UserControl
    WeightedSum
"""

from . import callback, control, debug, generators, mergers, noise, readers, writers
from .callback import CallbackComponent
from .control import TimeTrigger, UserControl
from .debug import DebugConsumer, DebugPushConsumer, ScheduleLogger
from .generators import CallbackGenerator, StaticCallbackGenerator
from .mergers import WeightedSum
from .noise import SimplexNoise, StaticSimplexNoise
from .parametric import ParametricGrid, StaticParametricGrid
from .readers import CsvReader
from .writers import CsvWriter

__all__ = [
    "callback",
    "control",
    "debug",
    "generators",
    "mergers",
    "noise",
    "readers",
    "writers",
]
__all__ += [
    "CallbackComponent",
    "CallbackGenerator",
    "CsvReader",
    "CsvWriter",
    "DebugConsumer",
    "DebugPushConsumer",
    "ParametricGrid",
    "ScheduleLogger",
    "SimplexNoise",
    "StaticCallbackGenerator",
    "StaticParametricGrid",
    "StaticSimplexNoise",
    "TimeTrigger",
    "UserControl",
    "WeightedSum",
]



"""
Modules for writing data.
"""
from datetime import datetime

import numpy as np

from ..data import tools as dtools
from ..data.grid_spec import NoGrid
from ..interfaces import ComponentStatus
from ..sdk import TimeComponent
from ..tools.date_helper import is_timedelta
from ..tools.log_helper import ErrorLogger


class CsvWriter(TimeComponent):
    """Writes CSV time series with one row per time step, from multiple inputs.

    Expects all inputs to be scalar values.

    .. code-block:: text

                     +-----------+
        --> [custom] |           |
        --> [custom] | CsvWriter |
        --> [......] |           |
                     +-----------+

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        writer = fm.components.CsvWriter(
            path="test.csv",
            inputs=["A", "B", "C"],
            time_column="T",
            separator=",",
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=1),
        )

    .. testcode:: constructor
        :hide:

        writer.initialize()

    Parameters
    ----------
    path : PathLike
        Path to the output file.
    inputs : list of str
        List of input names that will be written to file.
    start : :class:`datetime <datetime.datetime>`
        Starting time.
    step : :class:`timedelta <datetime.timedelta>` or :class:`relativedelta <dateutil.relativedelta.relativedelta>`
        Time step.
    time_column : str
        Time column name. Default "time"
    separator : str
        Column separator. Default ";"
    """

    def __init__(self, path, start, step, inputs, time_column="time", separator=";"):
        super().__init__()
        with ErrorLogger(self.logger):
            if not isinstance(start, datetime):
                raise ValueError("Start must be of type datetime")
            if not is_timedelta(step):
                raise ValueError("Step must be of type timedelta or relativedelta")

        self._path = path
        self._step = step
        self._time = start
        self._time_column = time_column
        self._separator = separator

        self._input_names = inputs

        self._rows = []

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        """Initialize the component.

        After the method call, the component's inputs and outputs must be available,
        and the component should have status INITIALIZED.
        """
        for inp in self._input_names:
            self.inputs.add(name=inp, time=None, grid=NoGrid(), units=None)

        self.create_connector(pull_data=self._input_names)

    def _connect(self, start_time):
        """Push initial values to outputs.

        After the method call, the component should have status CONNECTED.
        """
        self.try_connect(start_time)

        if self.status == ComponentStatus.CONNECTED:
            values = [
                dtools.get_magnitude(dtools.strip_time(data, NoGrid()))
                for _, data in self.connector.in_data.items()
            ]

            self._update_rows(values)

    def _validate(self):
        """Validate the correctness of the component's settings and coupling.

        After the method call, the component should have status VALIDATED.
        """

    def _update(self):
        """Update the component by one time step.
        Push new values to outputs.

        After the method call, the component should have status UPDATED or FINISHED.
        """
        self._time += self._step

        values = [
            dtools.get_magnitude(
                dtools.strip_time(self.inputs[inp].pull_data(self.time), NoGrid())
            )
            for inp in self._input_names
        ]
        self._update_rows(values)

    def _update_rows(self, values):
        with ErrorLogger(self.logger):
            for value, name in zip(values, self._input_names):
                dtools.assert_type(self, name, value.item(), [int, float])

        self._rows.append([self.time.isoformat()] + values)

    def _finalize(self):
        """Finalize and clean up the component.

        After the method call, the component should have status FINALIZED.
        """
        np.savetxt(
            self._path,
            self._rows,
            fmt="%s",
            delimiter=self._separator,
            header=self._separator.join([self._time_column] + self._input_names),
            comments="",
        )



"""Generic component with arbitrary inputs and extensive debug logging."""
import logging
from datetime import datetime, timedelta

from ..data import tools
from ..interfaces import ComponentStatus, IInput
from ..sdk import CallbackInput, Component, TimeComponent
from ..tools.date_helper import is_timedelta


class DebugConsumer(TimeComponent):
    """Generic component with arbitrary inputs and extensive debug logging.

    .. code-block:: text

                   +---------------+
      --> [custom] |               |
      --> [custom] | DebugConsumer |
      --> [......] |               |
                   +---------------+

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        component = fm.components.DebugConsumer(
            inputs={
                "A": fm.Info(time=None, grid=fm.NoGrid()),
                "B": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callbacks={
                "A": lambda n, d, t: print(t)
            },
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=7),
            log_data="INFO",
            strip_data=False,
        )

    .. testcode:: constructor
        :hide:

        component.initialize()



    Parameters
    ----------
    inputs : dict[str, Info]
        Dictionary of input names and infos.
    callbacks : dict[str, callable]
        Dictionary of optional input callbacks: callable(name, data, time).
    start : :class:`datetime <datetime.datetime>`
        Starting time
    step : :class:`timedelta <datetime.timedelta>` or :class:`relativedelta <dateutil.relativedelta.relativedelta>`
        Time step
    log_data : int or str or bool, optional
        Log level for printing received data, like "DEBUG" or "INFO".
        Default ``False``, logs nothing.

        ``True`` uses "INFO".
    strip_data : bool, optional
        Strips data before logging. Default ``True``.
    """

    def __init__(
        self, inputs, start, step, callbacks=None, log_data=False, strip_data=True
    ):
        super().__init__()

        if not isinstance(start, datetime):
            raise ValueError("Start must be of type datetime")
        if not is_timedelta(step):
            raise ValueError("Step must be of type timedelta or relativedelta")

        self._strip_data = strip_data
        self._log_data = None
        if isinstance(log_data, bool):
            if log_data:
                self._log_data = logging.INFO
        else:
            self._log_data = logging.getLevelName(log_data)

        self._input_infos = inputs
        self._callbacks = callbacks or {}
        self._step = step
        self._time = start
        self._data = {}

    @property
    def data(self):
        """dict[str, data] : The component's input data from the last time step"""
        return self._data

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        for name, info in self._input_infos.items():
            info.time = self.time
            self.inputs.add(name=name, info=info)
        self.create_connector(pull_data=list(self._input_infos.keys()))

    def _connect(self, start_time):
        self.try_connect(start_time)
        for name, info in self.connector.in_infos.items():
            if info is not None:
                self.logger.debug("Exchanged input info for %s", name)
        for name, data in self.connector.in_data.items():
            if data is not None:
                self.logger.debug("Pulled input data for %s", name)

                if self._log_data is not None:
                    pdata = (
                        tools.strip_time(data, self.inputs[name].info.grid)
                        if self._strip_data
                        else data
                    )
                    self.logger.log(
                        self._log_data,
                        'Received "%s" - %s: %s',
                        name,
                        self._time,
                        pdata,
                    )
                if name in self._callbacks:
                    self._callbacks[name](name, data, self._time)

                self._data[name] = data

    def _validate(self):
        pass

    def _update(self):
        self._time += self._step

        self._data = {
            n: self.inputs[n].pull_data(self.time) for n in self._input_infos.keys()
        }
        for name, data in self._data.items():
            if self._log_data is not None:
                pdata = (
                    tools.strip_time(data, self.inputs[name].info.grid)
                    if self._strip_data
                    else data
                )
                self.logger.log(
                    self._log_data,
                    'Received "%s" - %s: %s',
                    name,
                    self._time,
                    pdata,
                )
            if name in self._callbacks:
                self._callbacks[name](name, data, self._time)

    def _finalize(self):
        pass


class DebugPushConsumer(Component):
    """Generic component with arbitrary inputs and extensive debug logging. Push-based.

    .. code-block:: text

                   +-------------------+
      --> [custom] |                   |
      --> [custom] | DebugPushConsumer |
      --> [......] |                   |
                   +-------------------+

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        component = fm.components.DebugPushConsumer(
            inputs={
                "A": fm.Info(time=None, grid=fm.NoGrid()),
                "B": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callbacks={
                "A": lambda n, d, t: print(t)
            },
            log_data="INFO",
            strip_data=False,
        )

    .. testcode:: constructor
        :hide:

        component.initialize()



    Parameters
    ----------
    inputs : dict[str, Info]
        Dictionary of input names and infos.
    callbacks : dict[str, callable]
        Dictionary of optional input callbacks: callable(name, data, time).
    log_data : int or str or bool, optional
        Log level for printing received data, like "DEBUG" or "INFO".
        Default ``False``, logs nothing.

        ``True`` uses "INFO".
    strip_data : bool, optional
        Strips data before logging. Default ``True``.
    """

    def __init__(self, inputs, callbacks=None, log_data=False, strip_data=True):
        super().__init__()
        self._strip_data = strip_data
        self._log_data = None
        if isinstance(log_data, bool):
            if log_data:
                self._log_data = logging.INFO
        else:
            self._log_data = logging.getLevelName(log_data)

        self._input_infos = inputs
        self._callbacks = callbacks or {}
        self._data = {}

    @property
    def data(self):
        """dict[str, data] : The component's input data from the last time step"""
        return self._data

    def _initialize(self):
        for name, info in self._input_infos.items():
            self.inputs.add(
                CallbackInput(callback=self._data_pushed, name=name, info=info)
            )
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time)
        for name, info in self.connector.in_infos.items():
            if info is not None:
                self.logger.debug("Exchanged input info for %s", name)

    def _validate(self):
        pass

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _data_pushed(self, caller, time):
        data = caller.pull_data(time)
        self._data[caller.name] = data
        if self._log_data is not None:
            pdata = (
                tools.strip_time(data, caller.info.grid) if self._strip_data else data
            )
            self.logger.log(
                self._log_data,
                'Received "%s" - %s: %s',
                caller.name,
                time,
                pdata,
            )
        if caller.name in self._callbacks:
            self._callbacks[caller.name](caller.name, data, time)


class ScheduleLogger(Component):

    """Logging of module update schedule.

    Takes inputs of arbitrary types and simply logs the time of notifications of each input
    as an ASCII graph.

    .. code-block:: text

                     +----------------+
        --> [custom] |                |
        --> [custom] | ScheduleLogger |
        --> [......] |                |
                     +----------------+

    Note:
        This component is push-based without an internal time step.

    Examples
    --------

    .. testcode:: constructor

        from datetime import timedelta
        import finam as fm

        schedule = fm.components.ScheduleLogger(
            inputs={"Grid1": True, "Grid2": True},
            time_step=timedelta(days=1),
            log_level="DEBUG",
            stdout=True,
        )

    .. testcode:: constructor
        :hide:

        schedule.initialize()

    .. |relativedelta| replace:: :class:`relativedelta <dateutil.relativedelta.relativedelta>`

    Parameters
    ----------
    inputs : dict of str, bool
        Input names and whether to pull data from them when notified.
        Pulling is useful for correct output behaviour when clearing the data cache.
    time_step : :class:`timedelta <datetime.timedelta>` or |relativedelta|, optional
        Time per character in the ASCII graph. Default 1 day.
    log_level : str or int, optional
        Log level for the ASCII graph. Default "INFO".
    stdout : bool
        Prints the ASCII graphs to stdout.
        Useful for piping to file and/or for documentation
    """

    def __init__(
        self, inputs, time_step=timedelta(days=1), log_level="INFO", stdout=False
    ):
        super().__init__()
        self._pull_inputs = inputs
        self._time_step = time_step
        self._log_level = logging.getLevelName(log_level)
        self._stdout = stdout

        self._schedule = None
        self._output_map = None

    def _initialize(self):
        for inp in self._pull_inputs:
            self.inputs.add(
                CallbackInput(self._data_changed, name=inp, time=None, grid=None)
            )
        self.create_connector(
            pull_data=[inp for inp, pull in self._pull_inputs.items() if pull]
        )

    def _connect(self, start_time):
        self.try_connect(start_time)

    def _validate(self):
        pass

    def _data_changed(self, caller, time):
        self._update_schedule(caller, time)

    def _update(self):
        pass

    def _update_schedule(self, caller, time):
        if self._schedule is None:
            self._schedule = {inp: [] for _, inp in self.inputs.items()}
            self._output_map = {}
            for _, inp in self.inputs.items():
                out = inp
                while isinstance(out, IInput):
                    out = out.source
                self._output_map[inp] = out

        self._schedule[caller].append(time)

        if self._pull_inputs[caller.name]:
            _data = caller.pull_data(time)

        if self.status == ComponentStatus.VALIDATED:
            self._print_schedule(caller)

    def _print_schedule(self, caller):
        t_min = min(t[0] for _, t in self._schedule.items() if len(t) > 0)
        t_max = max(t[-1] for _, t in self._schedule.items() if len(t) > 0)
        t_diff = t_max - t_min

        num_char = int(t_diff / self._time_step) + 1

        self.logger.log(self._log_level, "input updated")
        if self._stdout:
            print("")

        max_name_len = max(len(inp.name) for inp in self._schedule)

        for inp, times in self._schedule.items():
            if len(times) == 0:
                continue

            s = [" "] * num_char

            out = self._output_map[inp]

            data_cache = getattr(out, "data", [])
            data_len = len(data_cache)
            i_min = len(times) - data_len

            prev = 0
            for i, t in enumerate(times):
                d = t - t_min
                pos = int(d / self._time_step)
                for j in range(prev, pos):
                    s[j] = "-"

                s[pos] = "o" if i >= i_min else "x"
                prev = pos + 1

            s = "".join(s)

            if inp == caller:
                s += " <-"

            s = f"{inp.name.ljust(max_name_len)} {s}"
            self.logger.log(self._log_level, s)
            if self._stdout:
                print(s)



"""Data tools for FINAM."""

from .core import (
    assert_type,
    check,
    full,
    full_like,
    has_time_axis,
    prepare,
    strip_time,
    to_datetime,
)
from .info import Info
from .mask import (
    Mask,
    check_data_covers_domain,
    filled,
    from_compressed,
    has_masked_values,
    is_masked_array,
    is_sub_mask,
    mask_specified,
    masks_compatible,
    masks_equal,
    to_compressed,
    to_masked,
)
from .units import (
    UNITS,
    check_quantified,
    clear_units_cache,
    compatible_units,
    equivalent_units,
    get_dimensionality,
    get_magnitude,
    get_units,
    is_quantified,
    quantify,
    to_units,
)

__all__ = [
    "UNITS",
    "Info",
    "Mask",
    "assert_type",
    "check",
    "check_data_covers_domain",
    "check_quantified",
    "clear_units_cache",
    "compatible_units",
    "equivalent_units",
    "filled",
    "from_compressed",
    "full",
    "full_like",
    "get_dimensionality",
    "get_magnitude",
    "get_units",
    "has_masked_values",
    "has_time_axis",
    "is_masked_array",
    "is_quantified",
    "is_sub_mask",
    "mask_specified",
    "masks_compatible",
    "masks_equal",
    "prepare",
    "quantify",
    "strip_time",
    "to_compressed",
    "to_datetime",
    "to_masked",
    "to_units",
]



"""Generator definitions."""

from datetime import datetime

from finam.interfaces import ComponentStatus

from ..sdk import Component, TimeComponent
from ..tools.date_helper import is_timedelta


class CallbackGenerator(TimeComponent):
    """Component to generate data in fixed time intervals from multiple callbacks.

    .. code-block:: text

        +-------------------+
        |                   | [custom] -->
        | CallbackGenerator | [custom] -->
        |                   | [......] -->
        +-------------------+

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        generator = fm.components.CallbackGenerator(
            callbacks={
                "Out1": (lambda t: t.day, fm.Info(time=None, grid=fm.NoGrid())),
                "Out2": (lambda t: t.month, fm.Info(time=None, grid=fm.NoGrid()))
            },
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=1)
        )

    .. testcode:: constructor
        :hide:

        generator.initialize()

    Parameters
    ----------
    callbacks : dict of (str, tuple(callable, Info))
        Dict of tuples (callback, info). ``callback(time) -> data`` per output name, returning the generated data.
    start : :class:`datetime <datetime.datetime>`
        Starting time.
    step : :class:`timedelta <datetime.timedelta>` or :class:`relativedelta <dateutil.relativedelta.relativedelta>`
        Time step.
    """

    def __init__(self, callbacks, start, step):
        super().__init__()

        if not isinstance(start, datetime):
            raise ValueError("Start must be of type datetime")
        if not is_timedelta(step):
            raise ValueError("Step must be of type timedelta or relativedelta")

        self._callbacks = callbacks
        self._step = step
        self._time = start
        self._initial_data = None

    def _next_time(self):
        return None

    def _initialize(self):
        """Initialize the component.

        After the method call, the component's inputs and outputs must be available,
        and the component should have status INITIALIZED.
        """
        for key, (_, info) in self._callbacks.items():
            info.time = self.time
            self.outputs.add(name=key, info=info)

        self.create_connector()

    def _connect(self, start_time):
        """Push initial values to outputs.

        After the method call, the component should have status CONNECTED.
        """
        if self._initial_data is None:
            self._initial_data = {
                key: callback(self._time)
                for key, (callback, _) in self._callbacks.items()
            }

        push_data = {}
        for name, req in self.connector.data_required.items():
            if req:
                push_data[name] = self._initial_data[name]

        self.try_connect(start_time, push_data=push_data)

        if self.status == ComponentStatus.CONNECTED:
            del self._initial_data
            del self._connector

    def _validate(self):
        """Validate the correctness of the component's settings and coupling.

        After the method call, the component should have status VALIDATED.
        """

    def _update(self):
        """Update the component by one time step.
        Push new values to outputs.

        After the method call, the component should have status UPDATED or FINISHED.
        """
        self._time += self._step

        for key, (callback, _) in self._callbacks.items():
            data = callback(self._time)
            if data is not None:
                self.outputs[key].push_data(data, self.time)

    def _finalize(self):
        """Finalize and clean up the component.

        After the method call, the component should have status FINALIZED.
        """


class StaticCallbackGenerator(Component):
    """Component to generate static data from multiple callbacks.

    .. code-block:: text

        +-------------------------+
        |                         | [custom] -->
        | StaticCallbackGenerator | [custom] -->
        |                         | [......] -->
        +-------------------------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        generator = fm.components.StaticCallbackGenerator(
            callbacks={
                "Out1": (lambda: 1.0, fm.Info(time=None, grid=fm.NoGrid())),
                "Out2": (lambda: 2.0, fm.Info(time=None, grid=fm.NoGrid()))
            },
        )

    .. testcode:: constructor
        :hide:

        generator.initialize()

    Parameters
    ----------
    callbacks : dict of (str, tuple(callable, Info))
        Dict of tuples (callback, info). ``callback() -> data`` per output name, returning the generated data.
    """

    def __init__(self, callbacks):
        super().__init__()
        self._callbacks = callbacks
        self._initial_data = None

    def _initialize(self):
        """Initialize the component.

        After the method call, the component's inputs and outputs must be available,
        and the component should have status INITIALIZED.
        """
        for key, (_, info) in self._callbacks.items():
            self.outputs.add(name=key, info=info, static=True)

        self.create_connector()

    def _connect(self, start_time):
        """Push initial values to outputs.

        After the method call, the component should have status CONNECTED.
        """
        if self._initial_data is None:
            self._initial_data = {
                key: callback() for key, (callback, _) in self._callbacks.items()
            }

        push_data = {}
        for name, pushed in self.connector.data_pushed.items():
            if not pushed:
                push_data[name] = self._initial_data[name]

        self.try_connect(start_time, push_data=push_data)

        if self.status == ComponentStatus.CONNECTED:
            del self._initial_data
            del self._connector

    def _validate(self):
        """Validate the correctness of the component's settings and coupling.

        After the method call, the component should have status VALIDATED.
        """

    def _update(self):
        """Update the component by one time step.
        Push new values to outputs.

        After the method call, the component should have status UPDATED or FINISHED.
        """

    def _finalize(self):
        """Finalize and clean up the component.

        After the method call, the component should have status FINALIZED.
        """



"""Module to provide unit support via pint approximating UDUNITS/CF.

Source: cf-xarray (https://github.com/xarray-contrib/cf-xarray/blob/main/cf_xarray/units.py)
License: Apache License 2.0 (https://github.com/xarray-contrib/cf-xarray/blob/main/LICENSE)
"""

# pylint: disable=W0212,W0613
import functools
import re
import warnings

import pint
from packaging.version import Version


@pint.register_unit_format("cf")
def short_formatter(unit, registry, **options):
    """Return a CF-compliant unit string from a `pint` unit.

    Parameters
    ----------
    unit : pint.UnitContainer
        Input unit.
    registry : pint.UnitRegistry
        The associated registry
    **options
        Additional options (may be ignored)

    Returns
    -------
    out : str
        Units following CF-Convention, using symbols.
    """
    # pint 0.24.1 gives {"dimensionless": 1} for non-shortened dimensionless units
    # CF uses "1" to denote fractions and dimensionless quantities
    if unit == {"dimensionless": 1} or not unit:
        return "1"

    # If u is a name, get its symbol (same as pint's "~" pre-formatter)
    # otherwise, assume a symbol (pint should have already raised on invalid units before this)
    unit = pint.util.UnitsContainer(
        {
            registry._get_symbol(u) if u in registry._units else u: exp
            for u, exp in unit.items()
        }
    )

    # Change in formatter signature in pint 0.24
    if Version(pint.__version__) < Version("0.24"):
        args = (unit.items(),)
    else:
        # Numerators splitted from denominators
        args = (
            ((u, e) for u, e in unit.items() if e >= 0),
            ((u, e) for u, e in unit.items() if e < 0),
        )

    out = pint.formatter(*args, as_ratio=False, product_fmt=" ", power_fmt="{}{}")
    # To avoid potentiel unicode problems in netCDF. In both cases, this unit is not recognized by udunits
    return out.replace("Δ°", "delta_deg")


# ------
# Reused with modification from MetPy under the terms of the BSD 3-Clause License.
# Copyright (c) 2015,2017,2019 MetPy Developers.
# Create registry, with preprocessors for UDUNITS-style powers (m2 s-2) and percent signs
units = pint.UnitRegistry(
    autoconvert_offset_to_baseunit=True,
    preprocessors=[
        functools.partial(
            re.compile(
                r"(?<=[A-Za-z])(?![A-Za-z])(?<![0-9\-][eE])(?<![0-9\-])(?=[0-9\-])"
            ).sub,
            "**",
        ),
        lambda string: string.replace("%", "percent"),
    ],
    force_ndarray_like=True,
)
# ----- end block copied from metpy

# need to insert to make sure this is the first preprocessor
# This ensures we convert integer `1` to string `"1"`, as needed by pint.
units.preprocessors.insert(0, str)

# -----
units.define("percent = 0.01 = %")

# Define commonly encountered units (both CF and non-CF) not defined by pint
units.define("@alias meter = gpm")
# ----- end block copied from metpy

# -----
# The following redefinitions were copied from xclim under the terms of their Apache-2 license
# In pint, the default symbol for year is "a" which is not CF-compliant (stands for "are")
units.define("year = 365.25 * day = yr")

# Define commonly encountered units not defined by pint
units.define("@alias degC = C = deg_C = Celsius = degrees_Celsius")
units.define("@alias degK = deg_K")
units.define("@alias day = d")
units.define("@alias hour = h")  # Not the Planck constant...
units.define(
    "degrees_north = degree = degrees_north = degrees_N = degreesN = degree_north = degree_N = degreeN"
)
units.define(
    "degrees_east = degree = degrees_east = degrees_E = degreesE = degree_east = degree_E = degreeE"
)
# degrees for grid_longitude / grid_latitude for grid_mappings
units.define("degrees = degree = degrees")
units.define("[speed] = [length] / [time]")
# ----- end block copied from xclim

# Add other specific aliases (by cf_xarray developers)
units.define("practical_salinity_unit = [] = psu = PSU")

# Enable pint's built-in matplotlib support
try:
    units.setup_matplotlib()
except ImportError:
    warnings.warn(
        "Import(s) unavailable to set up matplotlib support...skipping this portion "
        "of the setup."
    )
# end of vendored code from MetPy

# Set as application registry
pint.set_application_registry(units)



"""Iterative connection helpers."""
import copy
import logging
from abc import ABC

from finam.interfaces import ComponentStatus, Loggable

from ..data.tools import Info
from ..errors import FinamNoDataError, FinamTimeError
from ..tools.log_helper import ErrorLogger


class MissingInfoError(Exception):
    """Internal error type for handling missing infos for transfer rules"""


class InfoSource(ABC):
    """Base class for info transfer rules from inputs or outputs"""

    def __init__(self, name, fields):
        self.name = name

        if fields is not None and not isinstance(fields, list):
            raise TypeError(
                "Fields must be a list of metadata attributes as strings, or None."
            )

        self.fields = fields or []


class FromInput(InfoSource):
    """Info transfer rule from an input.

    See :meth:`.Component.create_connector` for usage details.

    Parameters
    ----------
    name : str
        Name of the input to take info from
    fields : list of str, optional
        Info fields to take from the input.
        Takes all fields if this is empty.
    """

    def __init__(self, name, fields=None):
        super().__init__(name, fields)


class FromOutput(InfoSource):
    """Info transfer rule from an output.

    See :meth:`.Component.create_connector` for usage details.

    Parameters
    ----------
    name : str
        Name of the output to take info from
    fields : list of str, optional
        Info fields to take from the output.
        Takes all fields if this is empty.
    """

    def __init__(self, name, fields=None):
        super().__init__(name, fields)


class FromValue:
    """
    Info transfer rule from a given value

    Parameters
    ----------
    field : str
        Field to set.
    value : any
        Value to set.
    """

    def __init__(self, field, value):
        self.field = field
        self.value = value


class ConnectHelper(Loggable):
    """Helper for iterative connect.

    Warning:
        This class is not intended for direct use!
        Use :meth:`.Component.create_connector` and :meth:`.Component.try_connect` instead.

    Parameters
    ----------
    logger : Logger
        Logger to use
    inputs : dict
        All inputs of the component.
    outputs : dict
        All outputs of the component.
    in_info_rules : dict
        Info transfer rules for inputs.
    out_info_rules : dict
        Info transfer rules for outputs.
    pull_data : arraylike
        Names of the inputs that are to be pulled.
    cache : bool
        Whether data and :class:`.Info` objects passed via :meth:`connect() <.ConnectHelper.connect>`
        are cached for later calls. Default ``True``.
    """

    def __init__(
        self,
        logger_name,
        inputs,
        outputs,
        pull_data=None,
        in_info_rules=None,
        out_info_rules=None,
        cache=True,
    ):
        super().__init__()

        self.base_logger_name = logger_name
        self._inputs = inputs
        self._outputs = outputs
        self._cache = cache

        with ErrorLogger(self.logger):
            for name in pull_data or []:
                if name not in self._inputs:
                    raise KeyError(
                        f"No input named '{name}' available to get info for."
                    )

        self._exchanged_in_infos = {name: None for name in self.inputs.keys()}
        self._exchanged_out_infos = {name: None for name in self.outputs.keys()}

        self._pulled_data = {name: None for name in pull_data or []}

        self._pushed_infos = {
            name: out.has_info() for name, out in self.outputs.items()
        }
        self._pushed_data = {
            name: False for name, out in self.outputs.items() if out.needs_push
        }

        self._in_info_rules = in_info_rules or {}
        self._out_info_rules = out_info_rules or {}

        with ErrorLogger(self.logger):
            self._check_info_rules()

        self._in_info_cache = {}
        self._out_info_cache = {}
        self._out_data_cache = {}

    def add_in_info_rule(self, in_name, rule):
        """
        Add an input info rule.

        Parameters
        ----------
        in_name : str
            Name of the input to add an info rule to.
        rule : FromOutput or FromInput or FromValue
            Rule to add.
        """
        if in_name in self._in_info_rules:
            self._in_info_rules[in_name].append(rule)
        else:
            self._in_info_rules[in_name] = [rule]
        with ErrorLogger(self.logger):
            self._check_info_rules()

    def add_out_info_rule(self, out_name, rule):
        """
        Add an output info rule.

        Parameters
        ----------
        out_name : str
            Name of the output to add an info rule to.
        rule : FromInput or FromOutput or FromValue
            Rule to add.
        """
        if out_name in self._out_info_rules:
            self._out_info_rules[out_name].append(rule)
        else:
            self._out_info_rules[out_name] = [rule]
        with ErrorLogger(self.logger):
            self._check_info_rules()

    def _apply_rules(self, rules):
        info = Info(time=None, grid=None)
        for rule in rules:
            if isinstance(rule, FromInput):
                in_info = self.in_infos[rule.name]
                if in_info is None:
                    raise MissingInfoError()
                _transfer_fields(in_info, info, rule.fields)
            elif isinstance(rule, FromOutput):
                out_info = self.out_infos[rule.name]
                if out_info is None:
                    raise MissingInfoError()
                _transfer_fields(out_info, info, rule.fields)
            elif isinstance(rule, FromValue):
                if rule.field == "time":
                    info.time = rule.value
                elif rule.field == "grid":
                    info.grid = rule.value
                else:
                    info.meta[rule.field] = rule.value

        return info

    def _check_info_rules(self):
        for name, rules in self._in_info_rules.items():
            if name not in self._inputs:
                raise KeyError(f"No input named '{name}' to apply info transfer rule.")
            for rule in rules:
                self._check_rule(rule)

        for name, rules in self._out_info_rules.items():
            if name not in self._outputs:
                raise KeyError(f"No output named '{name}' to apply info transfer rule.")
            for rule in rules:
                self._check_rule(rule)

    def _check_rule(self, rule):
        if isinstance(rule, FromInput):
            if rule.name not in self._inputs:
                raise KeyError(
                    f"No input named '{rule.name}' to use in info transfer rule."
                )
        elif isinstance(rule, FromOutput):
            if rule.name not in self._outputs:
                raise KeyError(
                    f"No output named '{rule.name}' to use in info transfer rule."
                )
        elif not isinstance(rule, FromValue):
            raise TypeError(
                f"Rules must be one of the types FromInput, FromOutput or FromValue. "
                f"Got '{rule.__class__.__name__}'."
            )

    @property
    def logger(self):
        """Logger for this component."""
        return logging.getLogger(self.logger_name)

    @property
    def logger_name(self):
        """Logger name derived from base logger name and class name."""
        base_logger = logging.getLogger(self.base_logger_name)
        # logger hierarchy indicated by "." in name
        return ".".join(([base_logger.name, self.__class__.__name__]))

    @property
    def uses_base_logger_name(self):
        """Whether this class has a ``base_logger_name`` attribute. True."""
        return True

    @property
    def inputs(self):
        """dict: The component's inputs."""
        return self._inputs

    @property
    def outputs(self):
        """dict: The component's outputs."""
        return self._outputs

    @property
    def in_infos(self):
        """dict: The exchanged input infos so far. May contain None values."""
        return self._exchanged_in_infos

    @property
    def out_infos(self):
        """dict: The exchanged output infos so far. May contain None values."""
        return self._exchanged_out_infos

    @property
    def in_data(self):
        """dict: The pulled input data so far. May contain None values."""
        return self._pulled_data

    @property
    def all_data_pulled(self):
        """bool: True if all expected data is pulled."""
        return all(data is not None for data in self.in_data.values())

    @property
    def infos_pushed(self):
        """dict: If an info was pushed for outputs so far."""
        return self._pushed_infos

    @property
    def data_pushed(self):
        """dict: If data was pushed for outputs so far."""
        return self._pushed_data

    @property
    def data_required(self):
        """dict: If data to push is still required."""
        return {
            name: not pushed and name not in self._out_data_cache
            for name, pushed in self.data_pushed.items()
        }

    @property
    def in_infos_required(self):
        """dict: If input infos to exchange are still required."""
        return {
            name: inf is None
            and self.inputs[name].info is None
            and name not in self._in_info_cache
            and name not in self._in_info_rules
            for name, inf in self.in_infos.items()
        }

    @property
    def out_infos_required(self):
        """dict: If output infos to push are still required."""
        return {
            name: not pushed
            and name not in self._out_info_cache
            and name not in self._out_info_rules
            for name, pushed in self.infos_pushed.items()
        }

    def connect(self, start_time, exchange_infos=None, push_infos=None, push_data=None):
        """Exchange the info and data with linked components.

        Values passed by the arguments are cached internally for later calls to the method
        if constructed with ``cache=True`` (the default).
        Thus, it is sufficient to provide only data and infos that became newly available.
        Giving the same data or infos repeatedly overwrites the cache.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`
            the composition's starting time as passed to :meth:`.Component.try_connect`
        exchange_infos : dict
            currently or newly available input data infos by input name
        push_infos : dict
            currently or newly available output data infos by output name
        push_data : dict
            currently or newly available output data by output name

        Returns
        -------
        ComponentStatus
            the new component status
        """

        exchange_infos = exchange_infos or {}
        push_infos = push_infos or {}
        push_data = push_data or {}

        with ErrorLogger(self.logger):
            self._check_names(exchange_infos, push_infos, push_data)
            self._check_in_rules(exchange_infos, push_infos)

        exchange_infos = {
            k: v for k, v in exchange_infos.items() if self.in_infos[k] is None
        }
        push_infos = {k: v for k, v in push_infos.items() if self.out_infos[k] is None}
        push_data = {k: v for k, v in push_data.items() if not self.data_pushed[k]}

        # Try to generate infos from transfer rules
        with ErrorLogger(self.logger):
            exchange_infos.update(self._apply_in_info_rules())
            push_infos.update(self._apply_out_info_rules())

        if self._cache:
            self._in_info_cache.update(exchange_infos)
            self._out_info_cache.update(push_infos)
            self._out_data_cache.update(push_data)
        else:
            self._in_info_cache = exchange_infos
            self._out_info_cache = push_infos
            self._out_data_cache = push_data

        any_done = self._exchange_in_infos()

        for name, info in self.out_infos.items():
            if info is None:
                try:
                    self.out_infos[name] = self.outputs[name].info
                    any_done = True
                    self.logger.trace("Successfully pulled output info for %s", name)
                except FinamNoDataError:
                    self.logger.trace("Failed to pull output info for %s", name)

        any_done += self._push(start_time)

        for name, data in self.in_data.items():
            if data is None:
                info = self.in_infos[name]
                if info is not None:
                    try:
                        self.in_data[name] = self.inputs[name].pull_data(
                            start_time or info.time
                        )
                        any_done = True
                        self.logger.trace("Successfully pulled input data for %s", name)
                    except FinamNoDataError:
                        self.logger.trace("Failed to pull input data for %s", name)

        if (
            all(v is not None for v in self.in_infos.values())
            and all(v is not None for v in self.out_infos.values())
            and all(v is not None for v in self.in_data.values())
            and all(v for v in self.infos_pushed.values())
            and all(v for v in self.data_pushed.values())
        ):
            with ErrorLogger(self.logger):
                _check_times(self.out_infos)

            return ComponentStatus.CONNECTED

        if any_done:
            return ComponentStatus.CONNECTING

        return ComponentStatus.CONNECTING_IDLE

    def _apply_in_info_rules(self):
        exchange_infos = {}
        for name, rules in self._in_info_rules.items():
            if self.in_infos[name] is None and name not in self._in_info_cache:
                try:
                    info = self._apply_rules(rules)
                    exchange_infos[name] = info
                except MissingInfoError:
                    pass
        return exchange_infos

    def _apply_out_info_rules(self):
        push_infos = {}
        for name, rules in self._out_info_rules.items():
            if not self.infos_pushed[name] and name not in self._out_info_cache:
                try:
                    info = self._apply_rules(rules)
                    push_infos[name] = info
                except MissingInfoError:
                    pass
        return push_infos

    def _check_names(self, exchange_infos, push_infos, push_data):
        for name in exchange_infos:
            if name not in self._inputs:
                raise KeyError(
                    f"No input named '{name}' available to exchange info for."
                )
        for name in push_infos:
            if name not in self._outputs:
                raise KeyError(f"No output named '{name}' available to push info.")
        for name in push_data:
            if name not in self._outputs:
                raise KeyError(f"No output named '{name}' available to push data.")

    def _check_in_rules(self, exchange_infos, push_infos):
        for name in exchange_infos:
            if name in self._in_info_rules:
                raise ValueError(
                    f"There are info transfer rules given for input `{name}`. "
                    f"Can't provide the info directly."
                )
        for name in push_infos:
            if name in self._out_info_rules:
                raise ValueError(
                    f"There are info transfer rules given for output `{name}`. "
                    f"Can't provide the info directly."
                )

    def _exchange_in_infos(self):
        any_done = False
        for name, info in self._exchanged_in_infos.items():
            if info is None and self.inputs[name].info is not None:
                try:
                    self.in_infos[name] = self.inputs[name].exchange_info()
                    any_done = True
                    self.logger.trace("Successfully exchanged input info for %s", name)
                except FinamNoDataError:
                    self.logger.trace("Failed to exchange input info for %s", name)

        for name, info in list(self._in_info_cache.items()):
            if self.in_infos[name] is None:
                try:
                    inf = self.inputs[name].info
                    self.in_infos[name] = self.inputs[name].exchange_info(
                        None if inf is not None else info
                    )
                    any_done = True
                    self._in_info_cache.pop(name)
                    self.logger.trace("Successfully exchanged input info for %s", name)
                except FinamNoDataError:
                    self.logger.trace("Failed to exchange input info for %s", name)

        return any_done

    def _push(self, time):
        any_done = False

        for name, info in list(self._out_info_cache.items()):
            if not self.infos_pushed[name]:
                self.outputs[name].push_info(info)
                self.infos_pushed[name] = True
                any_done = True
                self._out_info_cache.pop(name)
                self.logger.trace("Successfully pushed output info for %s", name)

        for name, data in list(self._out_data_cache.items()):
            if not self.data_pushed[name] and self.infos_pushed[name]:
                info = self.out_infos[name]
                if info is not None:
                    self._push_data(name, data, time, info.time)
                    any_done = True

        return any_done

    def _push_data(self, name, data, time, info_time):
        out = self.outputs[name]
        if out.is_static:
            out.push_data(data, None)
        elif info_time != time:
            out.push_data(data, time)
            out.push_data(copy.copy(data), info_time)
        else:
            out.push_data(data, info_time)

        self.data_pushed[name] = True
        self._out_data_cache.pop(name)
        self.logger.trace("Successfully pushed output data for %s", name)


def _check_times(infos):
    t = None
    for _, info in infos.items():
        if t is None:
            t = info.time
        elif t != info.time:
            raise FinamTimeError("Input infos have different starting times.")


def _transfer_fields(source_info, target_info, fields):
    if len(fields) == 0:
        target_info.time = source_info.time
        target_info.grid = source_info.grid
        target_info.meta = copy.copy(source_info.meta)
    else:
        for field in fields:
            if field == "time":
                target_info.time = source_info.time
            elif field == "grid":
                target_info.grid = source_info.grid
            else:
                target_info.meta[field] = source_info.meta[field]



"""Grid specifications to handle spatial data with FINAM."""

from pathlib import Path

import numpy as np
from pyevtk.hl import imageToVTK

from ..tools.enum_helper import get_enum_value
from .esri_tools import read_header
from .grid_base import Grid, GridBase, StructuredGrid
from .grid_tools import (
    CellType,
    Location,
    check_axes_monotonicity,
    gen_axes,
    prepare_vtk_data,
    prepare_vtk_kwargs,
)


def _check_location(grid, data_location):
    # need to define this here to prevent circular imports
    location = get_enum_value(data_location, Location)
    if location not in grid.valid_locations:
        msg = f"{grid.name}: data location {location} not valid."
        raise ValueError(msg)
    return location


class NoGrid(GridBase):
    """
    Indicator for data without a spatial grid.

    Parameters
    ----------
    dim : int or None, optional
        Data dimensionality. Should match the length of data_shape.
    data_shape : tuple of int or None, optional
        Data shape. Can contain -1 to indicate flexible axis.

    Raises
    ------
    ValueError
        If dim does not match the length of data_shape.
    """

    def __init__(self, dim=None, data_shape=None):
        if dim is None and data_shape is None:
            dim, data_shape = 0, tuple()
        if data_shape is None:
            data_shape = (-1,) * dim
        if dim is None:
            dim = len(data_shape)
        if dim != len(data_shape):
            msg = "NoGrid: dim needs to match the length of data_shape."
            raise ValueError(msg)
        self._dim = dim
        self._data_shape = data_shape

    @property
    def dim(self):
        """int: Dimension of the grid or data."""
        return self._dim

    @property
    def data_shape(self):
        """tuple: Shape of the associated data."""
        return self._data_shape

    # pylint: disable-next=unused-argument
    def compatible_with(self, other, check_location=True):
        """
        Check for compatibility with other Grid.

        Parameters
        ----------
        other : instance of Grid
            Other grid to compatibility with.
        check_location : bool, optional
            Whether to check location for equality, by default True

        Returns
        -------
        bool
            compatibility
        """
        return isinstance(other, NoGrid) and self.data_shape == other.data_shape

    def __eq__(self, other):
        return self.compatible_with(other)


class RectilinearGrid(StructuredGrid):
    """Regular grid with variable spacing in up to three coordinate directions.

    Parameters
    ----------
    axes : list of np.ndarray
        Axes defining the point coordinates in each direction (xyz order).
    data_location : Location, str, int, optional
        Data location in the grid, by default Location.CELLS
    order : str, optional
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"
    axes_reversed : bool, optional
        Indicate reversed axes order for the associated data, by default False
    axes_attributes : list of dict or None, optional
        Axes attributes following the CF convention (in xyz order), by default None
    axes_names : list of str or None, optional
        Axes names (in xyz order), by default ["x", "y", "z"]
    crs : str or None, optional
        The coordinate reference system, by default None
    """

    def __init__(
        self,
        axes,
        data_location=Location.CELLS,
        order="F",
        axes_reversed=False,
        axes_attributes=None,
        axes_names=None,
        crs=None,
    ):
        # at most 3 axes
        self._axes = [np.asarray(np.atleast_1d(ax), dtype=float) for ax in axes[:3]]
        # all axes made increasing
        self._axes_increase = check_axes_monotonicity(self.axes)
        self._dim = len(self.dims)
        self._data_location = None
        self.data_location = data_location
        self._order = order
        self._axes_reversed = bool(axes_reversed)
        self._axes_attributes = axes_attributes or (self.dim * [{}])
        if len(self.axes_attributes) != self.dim:
            raise ValueError("RectilinearGrid: wrong length of 'axes_attributes'")
        self._axes_names = axes_names or ["x", "y", "z"][: self.dim]
        if len(self.axes_names) != self.dim:
            raise ValueError("RectilinearGrid: wrong length of 'axes_names'")
        self._crs = crs

        self._data_shape = None
        self._data_size = None

    def to_unstructured(self):
        """
        Cast grid to an unstructured grid.

        Returns
        -------
        UnstructuredGrid
            Grid as unstructured grid.
        """
        return UnstructuredGrid(
            points=self.points,
            cells=self.cells,
            cell_types=self.cell_types,
            data_location=self.data_location,
            order=self.order,
            axes_attributes=self.axes_attributes,
            axes_names=self.axes_names,
            crs=self.crs,
        )

    @property
    def dims(self):
        """tuple: Axes lengths (xyz order)."""
        return tuple(map(len, self.axes))

    @property
    def data_shape(self):
        """tuple: Shape of the associated data."""
        if self._data_shape is None:
            self._data_shape = super().data_shape
        return self._data_shape

    @property
    def data_size(self):
        """int: Size of the associated data."""
        if self._data_size is None:
            self._data_size = super().data_size
        return self._data_size

    @property
    def axes(self):
        """list of np.ndarray: Grid points."""
        return self._axes

    @property
    def axes_reversed(self):
        """bool: Indicate reversed axes order for the associated data."""
        return self._axes_reversed

    @property
    def axes_increase(self):
        """list of bool: False to indicate a bottom up axis (xyz order)."""
        return self._axes_increase

    @property
    def axes_attributes(self):
        """list of dict: Axes attributes following the CF convention (xyz order)."""
        return self._axes_attributes

    @property
    def axes_names(self):
        """list of str: Axes names (xyz order)."""
        return self._axes_names

    @property
    def order(self):
        """str: Point, cell and data order (C- or Fortran-like)."""
        # vtk files use Fortran-like data ordering for structured grids
        return self._order

    @property
    def dim(self):
        """int: Dimension of the grid."""
        return self._dim

    @property
    def crs(self):
        """The coordinate reference system."""
        return self._crs

    @property
    def data_location(self):
        """Location of the associated data (either CELLS or POINTS)."""
        return self._data_location

    @data_location.setter
    def data_location(self, data_location):
        """Set location of the associated data (either CELLS or POINTS)."""
        self._data_location = _check_location(self, data_location)


class UniformGrid(RectilinearGrid):
    """Regular grid with uniform spacing in up to three coordinate directions.

    Parameters
    ----------
    dims : iterable
        Dimensions of the uniform grid for each direction.
        Spatial dimension will be determined by ``len(dims)``.
        Dimensions refer to the number of points, independent of ``data_location``.
    spacing : iterable, optional
        Spacing of the uniform in each dimension.  Defaults to
        ``(1.0, 1.0, 1.0)``. Must be positive.
    origin : iterable, optional
        Origin of the uniform grid.  Defaults to ``(0.0, 0.0, 0.0)``.
    data_location : Location, str, int, optional
        Data location in the grid, by default Location.CELLS
    order : str, optional
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "F"
    axes_reversed : bool, optional
        Indicate reversed axes order for the associated data, by default False
    axes_increase : arraylike or None, optional
        False to indicate a bottom up axis (xyz order), by default None
    axes_attributes : list of dict or None, optional
        Axes attributes following the CF convention (xyz order), by default None
    axes_names : list of str or None, optional
        Axes names (in xyz order), by default ["x", "y", "z"]
    crs : str or None, optional
        The coordinate reference system, by default None
    """

    def __init__(
        self,
        dims,
        spacing=(1.0, 1.0, 1.0),
        origin=(0.0, 0.0, 0.0),
        data_location=Location.CELLS,
        order="F",
        axes_reversed=False,
        axes_increase=None,
        axes_attributes=None,
        axes_names=None,
        crs=None,
    ):
        # at most 3 axes
        dims = tuple(dims)[:3]
        self.spacing = tuple(spacing)[: len(dims)]
        if len(self.spacing) < len(dims):
            raise ValueError("UniformGrid: wrong length of 'spacing'")
        self.origin = tuple(origin)[: len(dims)]
        if len(self.origin) < len(dims):
            raise ValueError("UniformGrid: wrong length of 'origin'")
        super().__init__(
            axes=gen_axes(dims, self.spacing, self.origin, axes_increase),
            data_location=data_location,
            order=order,
            axes_reversed=axes_reversed,
            axes_attributes=axes_attributes,
            axes_names=axes_names,
            crs=crs,
        )

    def export_vtk(
        self,
        path,
        data=None,
        cell_data=None,
        point_data=None,
        field_data=None,
        mesh_type="uniform",
    ):
        """
        Export grid and data to a VTK file.

        Parameters
        ----------
        path : pathlike
            File path.
            Suffix will be replaced according to mesh type (.vti, .vtr, .vtu)
        data : dict or None, optional
            Data in the corresponding shape given by name, by default None
        cell_data : dict or None, optional
            Additional cell data, by default None
        point_data : dict or None, optional
            Additional point data, by default None
        field_data : dict or None, optional
            Additional field data, by default None
        mesh_type : str, optional
            Mesh type ("uniform"/"structured"/"unstructured"),
            by default "structured"

        Raises
        ------
        ValueError
            If mesh type is not supported.
        """
        if mesh_type != "uniform":
            super().export_vtk(path, data, cell_data, point_data, field_data, mesh_type)
        else:
            data = prepare_vtk_data(data, self.axes_reversed, self.axes_increase)
            kw = prepare_vtk_kwargs(
                self.data_location, data, cell_data, point_data, field_data
            )
            path = str(Path(path).with_suffix(""))
            origin = self.origin + (0.0,) * (3 - self.dim)
            spacing = self.spacing + (0.0,) * (3 - self.dim)
            imageToVTK(path, origin, spacing, **kw)

    def to_rectilinear(self):
        """
        Cast grid to a rectilinear grid.

        Returns
        -------
        UniformGrid
            Grid as rectilinear grid.
        """
        grid = RectilinearGrid(
            axes=self.axes,
            data_location=self.data_location,
            order=self.order,
            axes_reversed=self.axes_reversed,
            axes_attributes=self.axes_attributes,
            axes_names=self.axes_names,
            crs=self.crs,
        )
        # pylint: disable-next=protected-access
        grid._axes_increase = self.axes_increase
        return grid


class EsriGrid(UniformGrid):
    """
    Esri grid raster specification.

    Parameters
    ----------
    ncols : int
        Number of grid columns, in cells.
    nrows : int
        Number of grid rows, in cells.
    cellsize : float, optional
        Cell size, by default 1.0
    xllcorner : float, optional
        x value of lower left corner, by default 0.0
    yllcorner : float, optional
        y value of lower left corner, by default 0.0
    order : str, optional
        Point and cell ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "C"
    axes_attributes : list of dict or None, optional
        Axes attributes following the CF convention (xyz order), by default None
    axes_names : list of str or None, optional
        Axes names (in xyz order), by default ["x", "y", "z"]
    crs : str or None, optional
        The coordinate reference system, by default None
    """

    valid_locations = (Location.CELLS,)
    """tuple: Valid locations for the grid."""

    def __init__(
        self,
        ncols,
        nrows,
        cellsize=1.0,
        xllcorner=0.0,
        yllcorner=0.0,
        order="C",
        axes_attributes=None,
        axes_names=None,
        crs=None,
    ):
        self.ncols = int(ncols)
        self.nrows = int(nrows)
        self.cellsize = float(cellsize)
        self.xllcorner = float(xllcorner)
        self.yllcorner = float(yllcorner)
        super().__init__(
            dims=(self.ncols + 1, self.nrows + 1),
            spacing=(self.cellsize, self.cellsize),
            origin=(self.xllcorner, self.yllcorner),
            order=order,
            axes_reversed=True,
            axes_increase=(True, False),
            axes_attributes=axes_attributes,
            axes_names=axes_names,
            crs=crs,
        )

    @classmethod
    def from_file(cls, file, axes_attributes=None, crs=None):
        """
        Generate EsriGrid from given file.

        Parameters
        ----------
        file : pathlike
            Path to the esri grid file
        axes_attributes : list of dict or None, optional
            Axes attributes following the CF convention (xyz order), by default None
        crs : str or None, optional
            The coordinate reference system, by default None

        Returns
        -------
        EsriGrid
            The grid specified in the file.
        """
        header = read_header(file)
        header.pop("nodata_value", None)
        header["crs"] = crs
        header["axes_attributes"] = axes_attributes
        return cls(**header)

    def to_uniform(self):
        """
        Cast grid to an uniform grid.

        Returns
        -------
        UniformGrid
            Grid as uniform grid.
        """
        return UniformGrid(
            dims=self.dims,
            spacing=self.spacing,
            origin=self.origin,
            data_location=self.data_location,
            order=self.order,
            axes_reversed=self.axes_reversed,
            axes_increase=self.axes_increase,
            axes_attributes=self.axes_attributes,
            axes_names=self.axes_names,
            crs=self.crs,
        )


class UnstructuredGrid(Grid):
    """
    Unstructured grid specification.

    Parameters
    ----------
    points : arraylike
        Points (n, dim) defining the grid.
    cells : arraylike
        Cells given by set list of point IDs defining the grid.
    cell_types : arraylike
        Cell types given as integer, e.g. CellType.TRI.
    data_location : Location, str, int, optional
        Data location in the grid, by default Location.CELLS
    order : str, optional
        Data ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "C"
    axes_attributes : list of dict or None, optional
        Axes attributes following the CF convention (in xyz order), by default None
    axes_names : list of str or None, optional
        Axes names (in xyz order), by default ["x", "y", "z"]
    crs : str or None, optional
        The coordinate reference system, by default None
    """

    def __init__(
        self,
        points,
        cells,
        cell_types,
        data_location=Location.CELLS,
        order="C",
        axes_attributes=None,
        axes_names=None,
        crs=None,
    ):
        # at most 3 axes
        self._points = np.asarray(np.atleast_2d(points), dtype=float)[:, :3]
        self._cells = np.asarray(np.atleast_2d(cells), dtype=int)
        self._cell_types = np.asarray(np.atleast_1d(cell_types), dtype=int)
        self._data_location = None
        self.data_location = data_location
        self._order = order
        self._axes_attributes = axes_attributes or (self.dim * [{}])
        if len(self.axes_attributes) != self.dim:
            raise ValueError("UnstructuredGrid: wrong length of 'axes_attributes'")
        self._axes_names = axes_names or ["x", "y", "z"][: self.dim]
        if len(self.axes_names) != self.dim:
            raise ValueError("UnstructuredGrid: wrong length of 'axes_names'")

        self._crs = crs

    @property
    def dim(self):
        """int: Dimension of the grid."""
        return self.points.shape[1]

    @property
    def crs(self):
        """The coordinate reference system."""
        return self._crs

    @property
    def point_count(self):
        """int: Number of grid points."""
        return len(self.points)

    @property
    def cell_count(self):
        """int: Number of grid cells."""
        return len(self.cells)

    @property
    def points(self):
        """np.ndarray: Grid points."""
        return self._points

    @property
    def data_shape(self):
        """np.ndarray: Grid points."""
        return (
            (len(self.points),)
            if self.data_location == Location.POINTS
            else (len(self.cells),)
        )

    @property
    def data_size(self):
        """int: Size of the associated data."""
        return (
            len(self.points)
            if self.data_location == Location.POINTS
            else len(self.cells)
        )

    @property
    def cells(self):
        """np.ndarray: Cell nodes in ESMF format."""
        return self._cells

    @property
    def cell_types(self):
        """np.ndarray: Cell types."""
        return self._cell_types

    @property
    def data_location(self):
        """Location of the associated data (either CELLS or POINTS)."""
        return self._data_location

    @data_location.setter
    def data_location(self, data_location):
        """Set location of the associated data (either CELLS or POINTS)."""
        self._data_location = _check_location(self, data_location)

    @property
    def order(self):
        """str: Point, cell and data order (C-like or F-like for flatten)."""
        return self._order

    @property
    def axes_attributes(self):
        """list of dict: Axes attributes following the CF convention (xyz order)."""
        return self._axes_attributes

    @property
    def axes_names(self):
        """list of str: Axes names (xyz order)."""
        return self._axes_names


class UnstructuredPoints(UnstructuredGrid):
    """
    Unstructured points without cells.

    Parameters
    ----------
    points : arraylike
        Points (n, dim) defining the grid.
    order : str, optional
        Data ordering.
        Either Fortran-like ("F") or C-like ("C"), by default "C"
    axes_attributes : list of dict or None, optional
        Axes attributes following the CF convention (in xyz order), by default None
    axes_names : list of str or None, optional
        Axes names (in xyz order), by default ["x", "y", "z"]
    crs : str or None, optional
        The coordinate reference system, by default None
    """

    valid_locations = (Location.POINTS,)
    """tuple: Valid locations for the grid."""

    def __init__(
        self,
        points,
        order="C",
        axes_attributes=None,
        axes_names=None,
        crs=None,
    ):
        # at most 3 axes
        pnt_cnt = len(points)
        super().__init__(
            points=points,
            cells=np.asarray([range(pnt_cnt)], dtype=int).T,
            cell_types=np.full(pnt_cnt, CellType.VERTEX, dtype=int),
            data_location=Location.POINTS,
            order=order,
            axes_attributes=axes_attributes,
            axes_names=axes_names,
            crs=crs,
        )

    @property
    def cell_centers(self):
        """np.ndarray: Grid cell centers."""
        return self.points

    @property
    def cell_node_counts(self):
        """np.ndarray: Node count for each cell."""
        return np.full(self.point_count, 1, dtype=int)

    @property
    def mesh_dim(self):
        """int: Maximal cell dimension."""
        return 0



import unittest
from datetime import datetime, timedelta

import finam as fm


class TestTimeTrigger(unittest.TestCase):
    def test_time_trigger(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15)),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 5))

    def test_time_trigger_from_source(self):
        time = datetime(2000, 1, 1)
        out_info = fm.Info(
            time=None,
            grid=fm.NoGrid(),
            units="m",
        )

        source = fm.components.CallbackGenerator(
            callbacks={
                "Out": (
                    lambda t: t.day,
                    fm.Info(time=None, grid=fm.NoGrid(), units="m"),
                ),
            },
            start=time,
            step=timedelta(days=1),
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            out_info=None,
            start=None,
            step=timedelta(days=1),
            start_from_input=True,
        )
        sink = fm.components.DebugConsumer(
            {"Input": out_info},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Out"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)

        self.assertEqual(trigger._time, datetime(2000, 1, 1))

        composition.run(end_time=datetime(2000, 1, 5))

    def test_time_trigger_from_source_with_info(self):
        time = datetime(2000, 1, 1)
        out_info = fm.Info(
            time=None,
            grid=fm.NoGrid(),
            units="m",
        )

        source = fm.components.CallbackGenerator(
            callbacks={
                "Out": (
                    lambda t: t.day,
                    fm.Info(time=None, grid=fm.NoGrid(), units="m"),
                ),
            },
            start=time,
            step=timedelta(days=1),
        )
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            out_info=fm.Info(time=None, grid=None, units=None),
            start=None,
            step=timedelta(days=1),
            start_from_input=True,
        )
        sink = fm.components.DebugConsumer(
            {"Input": out_info},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Out"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)

        self.assertEqual(trigger._time, datetime(2000, 1, 1))

        composition.run(end_time=datetime(2000, 1, 5))

    def test_time_trigger_from_target(self):
        time = datetime(2000, 1, 1)

        out_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15)),
            units="m",
        )

        source = fm.components.SimplexNoise()
        trigger = fm.components.TimeTrigger(
            out_info=fm.Info(time=None, grid=None, units=None),
            start=None,
            step=timedelta(days=1),
            start_from_input=False,
        )
        sink = fm.components.DebugConsumer(
            {"Input": out_info},
            start=time,
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)

        self.assertEqual(trigger._time, datetime(2000, 1, 1))

        composition.run(end_time=datetime(2000, 1, 5))

    def test_time_trigger_from_target_with_info(self):
        time = datetime(2000, 1, 1)

        out_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15)),
            units="m",
        )

        source = fm.components.SimplexNoise()
        trigger = fm.components.TimeTrigger(
            in_info=out_info.copy_with(),
            out_info=fm.Info(time=None, grid=None, units=None),
            start=None,
            step=timedelta(days=1),
            start_from_input=False,
        )
        sink = fm.components.DebugConsumer(
            {"Input": out_info},
            start=time,
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)

        self.assertEqual(trigger._time, datetime(2000, 1, 1))

        composition.run(end_time=datetime(2000, 1, 5))

    def test_time_trigger_fail(self):
        time = datetime(2000, 1, 1)

        trigger = fm.components.TimeTrigger(start=time, step=timedelta(days=1))
        with self.assertRaises(fm.FinamMetaDataError):
            trigger.initialize()

        trigger = fm.components.TimeTrigger(
            out_info=fm.Info(time=None, grid=None, units=None),
            start=None,
            step=timedelta(days=1),
            start_from_input=True,
        )
        with self.assertRaises(fm.FinamMetaDataError):
            trigger.initialize()

        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=None,
            step=timedelta(days=1),
            start_from_input=False,
        )
        with self.assertRaises(fm.FinamMetaDataError):
            trigger.initialize()


if __name__ == "__main__":
    unittest.main()



"""Date and time helper functions"""
from datetime import timedelta

try:
    from dateutil.relativedelta import relativedelta

    _DATEUTIL_INSTALLED = True
except ImportError:
    _DATEUTIL_INSTALLED = False


def is_timedelta(value):
    """
    Tests if a value is an instance of
    :class:`timedelta <datetime.timedelta>` or
    :class:`relativedelta <dateutil.relativedelta.relativedelta>`.
    """
    if isinstance(value, timedelta):
        return True

    if _DATEUTIL_INSTALLED and isinstance(value, relativedelta):
        return True

    return False



"""
Unit tests for masked data.
"""

import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm


def gen_masked(seed, shape):
    data = np.random.default_rng(seed).random(size=shape)
    mask = np.full(shape, False)
    # mask corners
    mask[0, 0] = mask[0, -1] = mask[-1, 0] = mask[-1, -1] = True
    return np.ma.masked_array(data, mask, fill_value=np.nan)


class TestMasked(unittest.TestCase):
    def test_rescale_masked(self):
        time = datetime(2000, 1, 1)

        in_info = fm.Info(
            time=time,
            grid=fm.UniformGrid(
                dims=(5, 4),
                spacing=(1.0, 1.0, 1.0),
                data_location=fm.Location.CELLS,
            ),
            units="m",
            missing_value=np.nan,
        )

        source = fm.components.generators.CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: gen_masked(t.toordinal(), in_info.grid.data_shape),
                    in_info,
                )
            },
            start=time,
            step=timedelta(days=1),
        )

        sink = fm.components.debug.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=time,
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, sink])
        source.outputs["Output"] >> fm.adapters.Scale(scale=2.0) >> sink.inputs["Input"]
        composition.connect()

        self.assertTrue(sink.data["Input"][0][0, 0].mask)
        self.assertAlmostEqual(
            sink.data["Input"][0][1, 1].magnitude,
            2 * source.outputs["Output"].data[0][1][0][1, 1].magnitude,
        )

    def test_mask_tools(self):
        data = gen_masked(1234, (4, 3))
        mask = data.mask
        xdata = fm.data.quantify(data)

        sub_domain = np.full_like(data, fill_value=False, dtype=bool)
        self.assertFalse(fm.data.check_data_covers_domain(data, sub_domain))
        self.assertFalse(fm.data.check_data_covers_domain(xdata, sub_domain))

        sub_domain = data.mask
        self.assertTrue(fm.data.check_data_covers_domain(data, sub_domain))
        self.assertTrue(fm.data.check_data_covers_domain(xdata, sub_domain))

        sub_domain[0, 1] = True
        self.assertTrue(fm.data.check_data_covers_domain(data, sub_domain))
        self.assertTrue(fm.data.check_data_covers_domain(xdata, sub_domain))

        with self.assertRaises(ValueError):
            self.assertTrue(fm.data.check_data_covers_domain(data, [False, True]))
            self.assertTrue(fm.data.check_data_covers_domain(xdata, [False, True]))

        fdata = fm.data.filled(data)
        fxdata = fm.data.filled(xdata)

        self.assertTrue(fm.data.is_masked_array(data))
        self.assertTrue(fm.data.is_masked_array(xdata))
        self.assertTrue(fm.data.has_masked_values(data))
        self.assertTrue(fm.data.has_masked_values(xdata))

        self.assertFalse(fm.data.is_masked_array(fdata))
        self.assertFalse(fm.data.is_masked_array(fxdata))
        self.assertFalse(fm.data.has_masked_values(fdata))
        self.assertFalse(fm.data.has_masked_values(fxdata))

        np.testing.assert_allclose(data[mask], fdata[mask])
        np.testing.assert_allclose(xdata[mask].magnitude, fxdata[mask].magnitude)

        mfdata = fm.data.to_masked(fdata, mask=False)
        mfxdata = fm.data.to_masked(fxdata, mask=False)

        self.assertTrue(fm.data.is_masked_array(mfdata))
        self.assertTrue(fm.data.is_masked_array(mfxdata))
        self.assertFalse(fm.data.has_masked_values(mfdata))
        self.assertFalse(fm.data.has_masked_values(mfxdata))

        cdata_c = fm.data.to_compressed(data, order="C")
        cdata_f = fm.data.to_compressed(data, order="F")
        ucdata_c = fm.data.from_compressed(
            cdata_c, shape=data.shape, order="C", mask=mask
        )
        ucdata_f = fm.data.from_compressed(
            cdata_f, shape=data.shape, order="F", mask=mask
        )

        self.assertEqual(cdata_c.size, np.sum(~mask))
        self.assertEqual(cdata_f.size, np.sum(~mask))
        self.assertNotEqual(cdata_c[1], cdata_f[1])

        np.testing.assert_allclose(data[mask], ucdata_c[mask])
        np.testing.assert_allclose(data[mask], ucdata_f[mask])

        # more specific routines
        grid1 = fm.RectilinearGrid([(1.0, 2.0, 3.0, 4.0)])
        grid2 = fm.RectilinearGrid([(1.0, 2.0, 3.0)])
        grid3 = fm.RectilinearGrid([(1.0, 2.0, 3.0), (1.0, 2.0, 3.0)])
        mask1 = np.array((1, 0, 0), dtype=bool)
        mask2 = np.array((0, 0, 1), dtype=bool)
        mask3 = np.array((1, 0, 1), dtype=bool)
        mask4 = np.array((1, 1, 1), dtype=bool)
        mask5 = np.array((0, 0, 0), dtype=bool)
        mask6 = np.array((1, 0), dtype=bool)
        mask7 = np.array(((0, 1), (0, 1)), dtype=bool)
        data = np.ma.masked_array((10.0, 20.0, 30.0), mask1, fill_value=np.nan)

        # submask check
        self.assertFalse(fm.data.tools.is_sub_mask(mask1, mask2))
        self.assertFalse(fm.data.tools.is_sub_mask(mask1, np.ma.nomask))
        self.assertTrue(fm.data.tools.is_sub_mask(mask1, mask3))
        self.assertTrue(fm.data.tools.is_sub_mask(mask2, mask3))
        self.assertTrue(fm.data.tools.is_sub_mask(np.ma.nomask, mask1))
        self.assertTrue(fm.data.tools.is_sub_mask(np.ma.nomask, mask2))
        self.assertTrue(fm.data.tools.is_sub_mask(np.ma.nomask, mask3))
        self.assertTrue(fm.data.tools.is_sub_mask(np.ma.nomask, mask4))
        self.assertTrue(fm.data.tools.is_sub_mask(np.ma.nomask, mask5))
        self.assertTrue(fm.data.tools.is_sub_mask(np.ma.nomask, np.ma.nomask))
        self.assertTrue(fm.data.tools.is_sub_mask(mask5, np.ma.nomask))
        self.assertFalse(fm.data.tools.is_sub_mask(mask1, mask6))
        self.assertFalse(fm.data.tools.is_sub_mask(mask1, mask7))

        # equal mask
        self.assertTrue(fm.data.tools.masks_equal(None, None))
        self.assertTrue(fm.data.tools.masks_equal(fm.Mask.NONE, fm.Mask.NONE))
        self.assertTrue(fm.data.tools.masks_equal(fm.Mask.FLEX, fm.Mask.FLEX))
        self.assertFalse(fm.data.tools.masks_equal(fm.Mask.FLEX, fm.Mask.NONE))
        self.assertTrue(fm.data.tools.masks_equal(np.ma.nomask, np.ma.nomask))
        self.assertTrue(fm.data.tools.masks_equal(np.ma.nomask, mask5))
        self.assertTrue(fm.data.tools.masks_equal(mask5, np.ma.nomask))
        self.assertFalse(fm.data.tools.masks_equal(mask1, mask6, grid1, grid2))
        self.assertFalse(fm.data.tools.masks_equal(mask1, mask7, grid1, grid3))
        self.assertFalse(fm.data.tools.masks_equal(mask1, mask2, grid1, grid1))

        # cover domain
        self.assertTrue(fm.data.tools.check_data_covers_domain(data, mask3))
        self.assertFalse(fm.data.tools.check_data_covers_domain(data, mask2))
        self.assertFalse(fm.data.tools.check_data_covers_domain(data, np.ma.nomask))

        np.testing.assert_array_almost_equal(data, fm.data.tools.to_masked(data))
        np.testing.assert_array_almost_equal((1, 2, 3), fm.data.tools.filled((1, 2, 3)))

    def test_info_mask(self):
        grid = fm.RectilinearGrid([(1.0, 2.0, 3.0)])
        mask = np.array((1, 0, 0), dtype=bool)
        with self.assertRaises(fm.FinamMetaDataError):
            fm.Info(grid=grid, mask=mask)


if __name__ == "__main__":
    unittest.main()



"""Units tools for FINAM."""

import numpy as np
import pint

from ...errors import FinamDataError

# pylint: disable-next=unused-import
from . import cf_units

# set default format to cf-convention for pint.dequantify
# some problems with degree_Celsius and similar here
pint.application_registry.default_format = "cf"
UNITS = pint.application_registry

_UNIT_PAIRS_CACHE = {}


def get_magnitude(xdata):
    """
    Get magnitude of given data.

    Parameters
    ----------
    xdata : pint.Quantity
        The given data array.

    Returns
    -------
    numpy.ndarray
        Magnitude of given data.
    """
    check_quantified(xdata, "get_magnitude")
    return xdata.magnitude


def get_units(xdata):
    """
    Get units of the data.

    Parameters
    ----------
    xdata : DataArray
        The given data array.

    Returns
    -------
    pint.Unit
        Units of the data.
    """
    check_quantified(xdata, "get_units")
    return xdata.units


def get_dimensionality(xdata):
    """
    Get dimensionality of the data.

    Parameters
    ----------
    xdata : pint.Quantity
        The given data array.

    Returns
    -------
    pint.UnitsContainer
        Dimensionality of the data.
    """
    check_quantified(xdata, "get_dimensionality")
    return xdata.dimensionality


def to_units(xdata, units, check_equivalent=False, report_conversion=False):
    """
    Convert data to given units.

    Parameters
    ----------
    xdata : pint.Quantity
        The given data array.
    units : str or pint.Unit
        Desired units.
    check_equivalent : bool, optional
        Checks for equivalent units and simply re-assigns if possible.
    report_conversion : bool, optional
        If true, returns a tuple with the second element indicating the unit conversion if it was required.

    Returns
    -------
    pint.Quantity or tuple(pint.Quantity, tuple(pint.Unit, pint.Unit) or None)
        The converted data.

        If ``report_conversion`` is ``True``, a tuple is returned with the second element
        indicating the unit conversion if it was required.

        The second element is ``None`` if no conversion was required,
        and a tuple of two :class:`pint.Unit` objects otherwise.
    """
    check_quantified(xdata, "to_units")
    units = _get_pint_units(units)
    units2 = xdata.units
    conversion = None
    if units != units2:
        if check_equivalent and equivalent_units(units, units2):
            xdata = UNITS.Quantity(xdata.magnitude, units)
        else:
            xdata = xdata.to(units)
            conversion = units2, units

    if report_conversion:
        return xdata, conversion
    return xdata


def is_quantified(xdata):
    """
    Check if data is a quantified DataArray.

    Parameters
    ----------
    xdata : Any
        The given data array.

    Returns
    -------
    bool
        Whether the data is a quantified DataArray.
    """
    return isinstance(xdata, pint.Quantity)


def quantify(xdata, units=None):
    """
    Quantifies data.

    Parameters
    ----------
    xdata : Any
        The given data array.
    units : UnitLike or Quantified or None, optional
        units to use, dimensionless by default

    Returns
    -------
    pint.Quantity
        The quantified array.
    """
    if is_quantified(xdata):
        raise FinamDataError(f"Data is already quantified with units '{xdata.units}'")
    return UNITS.Quantity(xdata, _get_pint_units(units or UNITS.dimensionless))


def check_quantified(xdata, routine="check_quantified"):
    """
    Check if data is a quantified DataArray.

    Parameters
    ----------
    xdata : numpy.ndarray
        The given data array.
    routine : str, optional
        Name of the routine to show in the Error, by default "check_quantified"

    Raises
    ------
    FinamDataError
        If the array is not a quantified DataArray.
    """
    if not is_quantified(xdata):
        raise FinamDataError(f"{routine}: given data is not quantified.")


def _get_pint_units(var):
    if var is None:
        raise FinamDataError("Can't extract units from 'None'.")

    if isinstance(var, pint.Unit):
        return var

    if isinstance(var, pint.Quantity):
        return var.units or UNITS.dimensionless

    return UNITS.Unit(var)


def compatible_units(unit1, unit2):
    """
    Checks if two units are compatible/convertible.

    Parameters
    ----------
    unit1 : UnitLike or Quantified
        First unit to compare.
    unit2 : UnitLike or Quantified
        Second unit to compare.

    Returns
    -------
    bool
        Unit compatibility.
    """
    unit1, unit2 = _get_pint_units(unit1), _get_pint_units(unit2)
    comp_equiv = _UNIT_PAIRS_CACHE.get((unit1, unit2))
    if comp_equiv is None:
        comp_equiv = _cache_units(unit1, unit2)

    return comp_equiv[0]


def equivalent_units(unit1, unit2):
    """
    Check if two given units are equivalent.

    Parameters
    ----------
    unit1 : UnitLike or Quantified
        First unit to compare.
    unit2 : UnitLike or Quantified
        Second unit to compare.

    Returns
    -------
    bool
        Unit equivalence.
    """
    unit1, unit2 = _get_pint_units(unit1), _get_pint_units(unit2)
    comp_equiv = _UNIT_PAIRS_CACHE.get((unit1, unit2))
    if comp_equiv is None:
        comp_equiv = _cache_units(unit1, unit2)

    return comp_equiv[1]


def _cache_units(unit1, unit2):
    equiv = False
    compat = False
    try:
        equiv = np.isclose((1.0 * unit1).to(unit2).magnitude, 1.0)
        compat = True
    except pint.errors.DimensionalityError:
        pass

    _UNIT_PAIRS_CACHE[(unit1, unit2)] = compat, equiv
    return compat, equiv


def clear_units_cache():
    """Clears the units cache"""
    _UNIT_PAIRS_CACHE.clear()



"""
Unit tests for the driver/scheduler.
"""

import logging
import os
import pprint
import unittest
from datetime import datetime, timedelta
from tempfile import TemporaryDirectory

import numpy as np

import finam as fm
from finam import (
    Adapter,
    CallbackInput,
    CallbackOutput,
    Component,
    ComponentStatus,
    Composition,
    FinamCircularCouplingError,
    FinamConnectError,
    FinamStatusError,
    Info,
    Input,
    NoBranchAdapter,
    NoGrid,
    Output,
    TimeComponent,
)
from finam._version import __version__
from finam.adapters.base import Scale
from finam.adapters.time import DelayFixed, NextTime
from finam.components import (
    CallbackComponent,
    CallbackGenerator,
    DebugPushConsumer,
    debug,
)
from finam.schedule import _check_dead_links, _find_dependencies


class NoTimeComponent(Component):
    def __init__(self):
        super().__init__()

    def _initialize(self):
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time)

    def _validate(self):
        pass

    def _update(self):
        pass

    def _finalize(self):
        pass


class MockupComponent(TimeComponent):
    def __init__(self, callbacks, step):
        super().__init__()

        if not isinstance(step, timedelta):
            raise ValueError("Step must be of type timedelta")

        self._callbacks = callbacks
        self._step = step
        self._time = datetime(2000, 1, 1)

        for key, _ in self._callbacks.items():
            self.outputs.add(name=key, time=self.time, grid=NoGrid())

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        self.create_connector()

    def _connect(self, start_time):
        push_data = {}
        for key, callback in self._callbacks.items():
            push_data[key] = callback(self._time)

        self.try_connect(start_time, push_data=push_data)

    def _validate(self):
        pass

    def _update(self):
        self._time += self._step

        for key, callback in self._callbacks.items():
            self.outputs[key].push_data(callback(self._time), self.time)

    def _finalize(self):
        pass


class MockupDependentComponent(TimeComponent):
    def __init__(self, step, static=False):
        super().__init__()
        self._step = step
        self._time = datetime(2000, 1, 1)
        self.static = static

        self.inputs.add(name="Input", static=self.static)

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        self.create_connector(pull_data=["Input"])

    def _connect(self, start_time):
        self.try_connect(
            start_time, exchange_infos={"Input": Info(time=self.time, grid=NoGrid())}
        )

    def _validate(self):
        pass

    def _update(self):
        _pulled = self.inputs["Input"].pull_data(self.time)
        self._time += self._step

    def _finalize(self):
        pass


class MockupCircularComponent(TimeComponent):
    def __init__(self, step):
        super().__init__()
        self._step = step
        self._time = datetime(2000, 1, 1)

        self.pulled_data = None

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        self.inputs.add(name="Input")
        self.outputs.add(name="Output", time=self.time, grid=NoGrid())
        self.create_connector(pull_data=["Input"])

    def _connect(self, start_time):
        push_data = {}
        pulled_data = self.connector.in_data["Input"]
        if pulled_data is not None:
            push_data["Output"] = pulled_data[0, ...]

        self.try_connect(
            start_time,
            exchange_infos={"Input": Info(time=self.time, grid=NoGrid())},
            push_data=push_data,
        )

    def _validate(self):
        pass

    def _update(self):
        self._time += self._step
        pulled = self.inputs["Input"].pull_data(self.time)
        self.outputs["Output"].push_data(pulled, self.time)

    def _finalize(self):
        pass


class CallbackAdapter(Adapter):
    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def _get_data(self, time):
        return self.callback(self.pull_data(time), time)


class NbAdapter(Adapter, NoBranchAdapter):
    def _get_data(self, time):
        return self.pull_data(time)


class TestComposition(unittest.TestCase):
    def test_init(self):
        module = MockupComponent(callbacks={"Output": lambda t: t}, step=timedelta(1.0))
        composition = Composition([module])

        self.assertEqual(module.status, ComponentStatus.INITIALIZED)
        self.assertEqual(len(module.outputs), 1)

    def test_check_composition(self):
        with self.assertRaises(ValueError):
            _comp = Composition(["not a component"])

    def test_validate_branching(self):
        module = MockupComponent(callbacks={"Output": lambda t: t}, step=timedelta(1.0))
        composition = Composition([module])

        non_branching_adapter = (
            module.outputs["Output"]
            >> NbAdapter()
            >> CallbackAdapter(callback=lambda data, time: data)
        )

        non_branching_adapter >> CallbackAdapter(callback=lambda data, time: data)
        non_branching_adapter >> CallbackAdapter(callback=lambda data, time: data)

        with self.assertRaises(FinamConnectError) as context:
            composition._validate_composition()

        self.assertTrue("Disallowed branching" in str(context.exception))

    def test_check_dead_links(self):
        info = Info(time=None, grid=NoGrid())

        out = Output(name="out", info=info)
        inp = Input(name="in", info=info)
        out >> inp

        self.assertTrue(inp.needs_pull)
        self.assertFalse(inp.needs_push)
        self.assertFalse(out.needs_pull)
        self.assertTrue(out.needs_push)

        _check_dead_links(0, inp)

        out = CallbackOutput(name="out", callback=None, info=info)
        inp = CallbackInput(name="in", callback=None, info=info)
        out >> inp

        self.assertFalse(inp.needs_pull)
        self.assertTrue(inp.needs_push)
        self.assertTrue(out.needs_pull)
        self.assertFalse(out.needs_push)

        with self.assertRaises(FinamConnectError):
            _check_dead_links(0, inp)

        out = Output(name="out", info=info)
        ada = NextTime()
        inp = CallbackInput(name="in", callback=None, info=info)

        out >> ada >> inp
        _check_dead_links(0, inp)

        out = CallbackOutput(name="out", callback=None, info=info)
        ada = NextTime()
        inp = Input(name="in", info=info)

        out >> ada >> inp
        with self.assertRaises(FinamConnectError):
            _check_dead_links(0, inp)

    def test_check_dead_links_error(self):
        info = Info(time=None, grid=NoGrid())
        out = CallbackOutput(name="out", callback=None, info=info)
        ada1 = Scale(2.0)
        ada2 = Scale(2.0)
        inp = CallbackInput(name="in", callback=None, info=info)
        out >> ada1 >> ada2 >> inp

        with self.assertRaises(FinamConnectError) as context:
            _check_dead_links(0, inp)

        message = str(context.exception)
        self.assertTrue("out >/> Scale >> Scale >/> in" in message)

    def test_validate_inputs(self):
        time = datetime(2000, 1, 1)
        module = debug.DebugConsumer(
            {"Input": Info(time=time, grid=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )
        composition = Composition([module])

        with self.assertRaises(FinamConnectError) as context:
            composition._validate_composition()

        self.assertTrue("Unconnected input" in str(context.exception))

    def test_log_file(self):
        with TemporaryDirectory() as tmp:
            log_file = os.path.join(tmp, "test.log")

            module1 = MockupComponent(
                callbacks={"Output": lambda t: t.day}, step=timedelta(1.0)
            )
            module2 = MockupDependentComponent(step=timedelta(1.0))

            composition = Composition(
                [module2, module1], log_level=logging.DEBUG, log_file=log_file
            )

            module1.outputs["Output"] >> module2.inputs["Input"]

            composition.run(
                start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 2)
            )

            with open(log_file) as f:
                lines = f.readlines()
                self.assertNotEqual(len(lines), 0)

    def test_collect_adapters(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: t.day}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module2, module1])

        ada = fm.adapters.Scale(1.0)
        module1.outputs["Output"] >> ada >> module2.inputs["Input"]

        composition.connect()

        self.assertEqual({ada}, composition._adapters)

    def test_fail_time(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: t}, step=timedelta(1.0)
        )
        composition = Composition([module1])

        with self.assertRaises(ValueError):
            composition.run(start_time=0, end_time=100)

    def test_fail_double_connect(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: t}, step=timedelta(1.0)
        )
        composition = Composition([module1])
        composition.connect(datetime(2000, 1, 1))

        with self.assertRaises(FinamStatusError):
            composition.connect(datetime(2000, 1, 1))

    def test_base_logger(self):
        composition = Composition([])
        self.assertFalse(composition.uses_base_logger_name)

    def test_iterative_connect(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: t.day}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module2, module1])

        module1.outputs["Output"] >> module2.inputs["Input"]

        composition.run(start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 31))

    def test_iterative_connect_multi(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: t.day}, step=timedelta(1.0)
        )
        module2 = MockupCircularComponent(step=timedelta(1.0))
        module3 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module3, module2, module1])

        module1.outputs["Output"] >> module2.inputs["Input"]
        module2.outputs["Output"] >> module3.inputs["Input"]

        composition.run(start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 31))

    def test_iterative_connect_adapter(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module2, module1])

        module1.outputs["Output"] >> Scale(1.0) >> module2.inputs["Input"]

        composition.run(start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 31))

    def test_iterative_connect_multi_adapter(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module2, module1])

        module1.outputs["Output"] >> Scale(1.0) >> Scale(1.0) >> module2.inputs["Input"]

        composition.run(start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 31))

    def test_iterative_connect_blocked(self):
        module1 = MockupCircularComponent(step=timedelta(1.0))
        module2 = MockupCircularComponent(step=timedelta(1.0))

        composition = Composition([module1, module2])

        module1.outputs["Output"] >> module2.inputs["Input"]
        module2.outputs["Output"] >> module1.inputs["Input"]

        with self.assertRaises(FinamCircularCouplingError):
            composition.run(
                start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 31)
            )

    def test_no_time_comp(self):
        module = NoTimeComponent()

        composition = Composition([module])

        with self.assertRaises(ValueError):
            composition.run(end_time=datetime(2000, 1, 31))

        composition.run(end_time=None)

    def test_time_comp(self):
        module = MockupComponent(callbacks={"Output": lambda t: t}, step=timedelta(1.0))

        composition = Composition([module])

        with self.assertRaises(ValueError):
            composition.run(end_time=None)

        composition.run(start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 2))

    def test_no_update(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(days=1)
        )
        module2 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(days=31)
        )
        composition = Composition([module1, module2])
        composition.run(start_time=datetime(2000, 1, 1), end_time=datetime(2000, 1, 1))

    def test_missing_component_upstream(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module2])

        module1.outputs["Output"] >> Scale(1.0) >> Scale(1.0) >> module2.inputs["Input"]

        with self.assertRaises(FinamConnectError):
            composition.connect(
                datetime(2000, 1, 1),
            )

    def test_missing_component_downstream(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module1])

        module1.outputs["Output"] >> Scale(1.0) >> Scale(1.0) >> module2.inputs["Input"]

        with self.assertRaises(FinamConnectError):
            composition.connect(datetime(2000, 1, 1))

    def test_dependencies_simple(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module1, module2])

        module1.outputs["Output"] >> Scale(1.0) >> module2.inputs["Input"]

        composition.connect(datetime(2000, 1, 1))

        self.assertEqual(
            _find_dependencies(
                module1, composition._output_owners, datetime(2000, 1, 5)
            ),
            {},
        )
        self.assertEqual(
            _find_dependencies(
                module2, composition._output_owners, datetime(2000, 1, 1)
            ),
            {},
        )
        self.assertEqual(
            _find_dependencies(
                module2, composition._output_owners, datetime(2000, 1, 5)
            ),
            {module1.outputs["Output"]: (datetime(2000, 1, 5), False)},
        )

    def test_dependencies_multi(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: 1.0}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))

        module3 = MockupCircularComponent(step=timedelta(1.0))
        module4 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module1, module2, module3, module4])

        module1.outputs["Output"] >> Scale(1.0) >> module2.inputs["Input"]
        module1.outputs["Output"] >> Scale(1.0) >> module3.inputs["Input"]
        (
            module3.outputs["Output"]
            >> Scale(1.0)
            >> DelayFixed(timedelta(days=2))
            >> module4.inputs["Input"]
        )

        composition.connect(datetime(2000, 1, 1))

        self.assertEqual(
            _find_dependencies(
                module1, composition._output_owners, datetime(2000, 1, 5)
            ),
            {},
        )
        self.assertEqual(
            _find_dependencies(
                module2, composition._output_owners, datetime(2000, 1, 5)
            ),
            {module1.outputs["Output"]: (datetime(2000, 1, 5), False)},
        )
        self.assertEqual(
            _find_dependencies(
                module3, composition._output_owners, datetime(2000, 1, 5)
            ),
            {module1.outputs["Output"]: (datetime(2000, 1, 5), False)},
        )

        self.assertEqual(
            _find_dependencies(
                module4, composition._output_owners, datetime(2000, 1, 1)
            ),
            {},
        )
        self.assertEqual(
            _find_dependencies(
                module4, composition._output_owners, datetime(2000, 1, 2)
            ),
            {},
        )
        self.assertEqual(
            _find_dependencies(
                module4, composition._output_owners, datetime(2000, 1, 5)
            ),
            {module3.outputs["Output"]: (datetime(2000, 1, 3), True)},
        )

    def test_static_run(self):
        info = fm.Info(time=None, grid=fm.NoGrid())

        source = fm.components.StaticSimplexNoise(info=info, seed=123)
        sink = fm.components.DebugPushConsumer(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
        )

        composition = Composition([source, sink])

        source.outputs["Noise"] >> Scale(1.0) >> sink.inputs["In"]

        with self.assertRaises(ValueError):
            composition.run(end_time=datetime(2000, 1, 1))

        composition.run()

        # We get data without the time dimension here
        self.assertEqual(sink.data["In"].shape, (1,))

    def test_static_fail(self):
        info = fm.Info(time=None, grid=fm.NoGrid())

        source = fm.components.SimplexNoise(info=info, seed=123)
        sink = MockupDependentComponent(step=timedelta(days=1), static=True)

        composition = Composition([source, sink])

        source.outputs["Noise"] >> Scale(1.0) >> sink.inputs["Input"]

        with self.assertRaises(FinamConnectError):
            composition.connect(datetime(2000, 1, 1))

    def test_dependencies_schedule(self):
        start = datetime(2000, 1, 1)
        info = fm.Info(time=start, grid=fm.NoGrid())

        updates = []

        def lambda_generator(t):
            updates.append("A")
            return t.day

        def lambda_component(inp, _t):
            updates.append("B")
            return {"Out": inp["In"]}

        def lambda_component_2(inp, _t):
            updates.append("C")
            return {"Out": inp["In"]}

        module1 = CallbackGenerator(
            callbacks={"Out": (lambda_generator, info)},
            start=start,
            step=timedelta(1.0),
        )
        module2 = CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda_component,
            start=start,
            step=timedelta(days=5),
        )
        module3 = CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda_component_2,
            start=start,
            step=timedelta(days=8),
        )
        composition = Composition([module1, module2, module3])

        module1.outputs["Out"] >> Scale(1.0) >> module2.inputs["In"]
        module2.outputs["Out"] >> Scale(1.0) >> module3.inputs["In"]

        composition.connect(datetime(2000, 1, 1))
        self.assertEqual(updates, ["A", "B", "C"])

        composition.run(end_time=datetime(2000, 1, 2))
        self.assertEqual(
            updates,
            [
                "A",
                "B",
                "C",  # Connect
                "A",
                "A",
                "A",
                "A",
                "A",
                "B",  # Update B to 5
                "A",
                "A",
                "A",
                "A",
                "A",
                "B",  # Update B to 10
                "C",  # Update C to 8
            ],
        )

    def test_dependencies_schedule_no_push(self):
        start = datetime(2000, 1, 1)
        info = fm.Info(time=start, grid=fm.NoGrid())

        updates = []

        def lambda_generator(t):
            updates.append("A")
            if (t.day - 1) % 5 != 0:
                return None
            return t.day

        def lambda_component(inp, t):
            updates.append("B")
            return {"Out": inp["In"]}

        module1 = CallbackGenerator(
            callbacks={"Out": (lambda_generator, info)},
            start=start,
            step=timedelta(days=1),
        )
        module2 = CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda_component,
            start=start,
            step=timedelta(days=1),
        )
        composition = Composition([module1, module2])

        module1.outputs["Out"] >> Scale(1.0) >> module2.inputs["In"]

        composition.connect()
        self.assertEqual(updates, ["A", "B"])

        composition.run(end_time=datetime(2000, 1, 2))
        self.assertEqual(
            updates,
            [
                "A",
                "B",  # Connect
                "A",
                "A",
                "A",
                "A",
                "A",
                "B",  # Update B to 5
            ],
        )

    def test_dependencies_schedule_no_push_double(self):
        start = datetime(2000, 1, 1)
        info = fm.Info(time=start, grid=fm.NoGrid())

        updates = []

        def lambda_generator_1(t):
            updates.append("A1")
            return t.day

        def lambda_generator_2(t):
            if (t.day - 1) % 5 != 0:
                return None
            updates.append("A2")
            return t.day

        def lambda_component(inp, t):
            updates.append("B")
            return {"Out": inp["In1"]}

        module1 = CallbackGenerator(
            callbacks={
                "Out1": (lambda_generator_1, info),
                "Out2": (lambda_generator_2, info),
            },
            start=start,
            step=timedelta(days=1),
        )
        module2 = CallbackComponent(
            inputs={
                "In1": fm.Info(time=None, grid=fm.NoGrid()),
                "In2": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda_component,
            start=start,
            step=timedelta(days=1),
        )
        composition = Composition([module1, module2])

        module1.outputs["Out1"] >> Scale(1.0) >> module2.inputs["In1"]
        module1.outputs["Out2"] >> Scale(1.0) >> module2.inputs["In2"]

        composition.connect()
        self.assertEqual(updates, ["A1", "A2", "B"])

        composition.run(end_time=datetime(2000, 1, 2))
        self.assertEqual(
            updates,
            [
                "A1",
                "A2",
                "B",  # Connect
                "A1",
                "A1",
                "A1",
                "A1",
                "A1",
                "A2",
                "B",  # Update B to 5
            ],
        )

    def test_dependency_fail(self):
        start = datetime(2000, 1, 1)
        info = fm.Info(time=start, grid=fm.NoGrid())

        updates = []

        def lambda_component(inp, _t):
            return {"Out": inp["In"] if inp else 0.0}

        module1 = CallbackComponent(
            inputs={"In": info},
            outputs={"Out": info},
            callback=lambda_component,
            start=start,
            step=timedelta(days=5),
        )
        module2 = CallbackComponent(
            inputs={"In": info},
            outputs={"Out": info},
            callback=lambda_component,
            start=start,
            step=timedelta(days=8),
            initial_pull=False,
        )
        composition = Composition([module1, module2])

        module1.outputs["Out"] >> Scale(1.0) >> module2.inputs["In"]
        module2.outputs["Out"] >> Scale(1.0) >> module1.inputs["In"]

        composition.connect(start)

        with self.assertRaises(FinamCircularCouplingError):
            composition.run(end_time=datetime(2000, 1, 2))

    def test_starting_time(self):
        start_1 = datetime(2000, 1, 1)
        start_2 = datetime(2000, 1, 8)

        updates = {"A": [], "B": []}

        def lambda_generator(t):
            return t.day

        def lambda_component(inp, t):
            return {"Out": inp["In"][0, ...]}

        def lambda_debugger(name, data, t):
            updates[name].append(t.day)

        module1 = CallbackGenerator(
            callbacks={"Out": (lambda_generator, fm.Info(time=None, grid=fm.NoGrid()))},
            start=start_2,
            step=timedelta(days=5),
        )
        module2 = CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda_component,
            start=start_1,
            step=timedelta(days=3),
        )
        module3 = DebugPushConsumer(
            inputs={
                "A": fm.Info(time=None, grid=None),
                "B": fm.Info(time=None, grid=None),
            },
            callbacks={
                "A": lambda_debugger,
                "B": lambda_debugger,
            },
        )

        composition = Composition([module1, module2, module3])

        module1.outputs["Out"] >> Scale(1.0) >> module2.inputs["In"]
        module1.outputs["Out"] >> Scale(1.0) >> module3.inputs["A"]
        module2.outputs["Out"] >> Scale(1.0) >> module3.inputs["B"]

        composition.connect()

        composition.run(end_time=datetime(2000, 1, 10))

        self.assertEqual([1, 8, 13], updates["A"])
        self.assertEqual([1, 4, 7, 10], updates["B"])

    def test_starting_time(self):
        start_1 = datetime(2000, 1, 2)
        start_2 = datetime(2000, 1, 8)

        updates = {"A": [], "B": []}

        def lambda_generator(t):
            return t.day

        def lambda_component(inp, t):
            return {"Out": inp["In"].copy()}

        def lambda_debugger(name, data, t):
            updates[name].append(t.day)

        module1 = CallbackGenerator(
            callbacks={"Out": (lambda_generator, fm.Info(time=None, grid=fm.NoGrid()))},
            start=start_2,
            step=timedelta(days=5),
        )
        module2 = CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda_component,
            start=start_1,
            step=timedelta(days=3),
        )
        module3 = DebugPushConsumer(
            inputs={
                "A": fm.Info(time=None, grid=None),
                "B": fm.Info(time=None, grid=None),
            },
            callbacks={
                "A": lambda_debugger,
                "B": lambda_debugger,
            },
        )

        composition = Composition([module1, module2, module3])

        module1.outputs["Out"] >> Scale(1.0) >> module2.inputs["In"]
        module1.outputs["Out"] >> Scale(1.0) >> module3.inputs["A"]
        module2.outputs["Out"] >> Scale(1.0) >> module3.inputs["B"]

        composition.connect(datetime(2000, 1, 1))

        composition.run(end_time=datetime(2000, 1, 10))

        self.assertEqual([1, 8, 13], updates["A"])
        self.assertEqual([1, 2, 5, 8, 11], updates["B"])

    def test_metadata(self):
        module1 = MockupComponent(
            callbacks={"Output": lambda t: t.day}, step=timedelta(1.0)
        )
        module2 = MockupDependentComponent(step=timedelta(1.0))
        module3 = MockupDependentComponent(step=timedelta(1.0))

        composition = Composition([module2, module1, module3])

        ada1 = fm.adapters.Scale(1.0)
        ada2 = fm.adapters.Scale(1.0)
        module1.outputs["Output"] >> ada1 >> ada2 >> module2.inputs["Input"]
        module1.outputs["Output"] >> module3.inputs["Input"]

        with self.assertRaises(FinamStatusError) as context:
            _ = composition.metadata

        composition.connect()

        md = composition.metadata

        pprint.pprint(md)

        self.assertIn("components", md)
        self.assertIn("adapters", md)
        self.assertIn("links", md)

        self.assertEqual([datetime(2000, 1, 1), None], md["time_frame"])

        self.assertIn(f"{module1.name}@{id(module1)}", md["components"])
        self.assertIn(f"{module2.name}@{id(module2)}", md["components"])
        self.assertIn(f"{ada1.name}@{id(ada1)}", md["adapters"])
        self.assertIn(f"{ada2.name}@{id(ada2)}", md["adapters"])

        self.assertEqual(4, len(md["links"]))
        self.assertTrue(
            {
                "from": {
                    "component": f"{module1.name}@{id(module1)}",
                    "output": "Output",
                },
                "to": {"component": f"{module3.name}@{id(module3)}", "input": "Input"},
            }
            in md["links"]
        )
        self.assertTrue(
            {
                "from": {
                    "component": f"{module1.name}@{id(module1)}",
                    "output": "Output",
                },
                "to": {"adapter": f"{ada1.name}@{id(ada1)}"},
            }
            in md["links"]
        )
        self.assertTrue(
            {
                "from": {"adapter": f"{ada1.name}@{id(ada1)}"},
                "to": {"adapter": f"{ada2.name}@{id(ada2)}"},
            }
            in md["links"]
        )
        self.assertTrue(
            {
                "from": {"adapter": f"{ada2.name}@{id(ada2)}"},
                "to": {"component": f"{module2.name}@{id(module2)}", "input": "Input"},
            }
            in md["links"]
        )

        self.assertEqual(__version__, md["version"])


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm
import finam.errors
from finam.components.debug import DebugConsumer
from finam.components.generators import CallbackGenerator
from finam.components.mergers import WeightedSum


def generate_grid(grid):
    return np.reshape(
        np.random.random(grid.data_size), newshape=grid.data_shape, order=grid.order
    )


class TestWeightedSum(unittest.TestCase):
    def test_weighted_sum_simple(self):
        start = datetime(2000, 1, 1)

        generator1 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: 1.0,
                    fm.Info(None, grid=fm.NoGrid(), units="mm"),
                ),
                "Weight": (lambda t: 0.25, fm.Info(None, grid=fm.NoGrid())),
            },
            start=start,
            step=timedelta(days=1),
        )
        generator2 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: 2.0,
                    fm.Info(None, grid=fm.NoGrid(), units="mm"),
                ),
                "Weight": (lambda t: 0.75, fm.Info(None, grid=fm.NoGrid())),
            },
            start=start,
            step=timedelta(days=5),
        )

        merger = WeightedSum(inputs=["In1", "In2"])

        consumer = DebugConsumer(
            inputs={"WeightedSum": fm.Info(None, grid=None, units=None)},
            start=start,
            step=timedelta(days=1),
        )

        composition = fm.Composition([generator1, generator2, merger, consumer])

        generator1.outputs["Value"] >> merger.inputs["In1"]
        generator1.outputs["Weight"] >> merger.inputs["In1_weight"]

        generator2.outputs["Value"] >> merger.inputs["In2"]
        generator2.outputs["Weight"] >> merger.inputs["In2_weight"]

        merger.outputs["WeightedSum"] >> consumer.inputs["WeightedSum"]

        composition.run(start_time=start, end_time=start + timedelta(days=30))

        self.assertEqual(consumer.data["WeightedSum"], 1.75 * fm.UNITS.millimeter)
        self.assertEqual(
            fm.data.get_units(consumer.data["WeightedSum"]), fm.UNITS("mm")
        )

    def test_weighted_sum(self):
        start = datetime(2000, 1, 1)
        source_grid = fm.UniformGrid((20, 15))

        generator1 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=1),
        )
        generator2 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=5),
        )

        merger = WeightedSum(inputs=["In1", "In2"])

        consumer = DebugConsumer(
            inputs={"WeightedSum": fm.Info(None, grid=None, units=None)},
            start=start,
            step=timedelta(days=1),
        )

        composition = fm.Composition([generator1, generator2, merger, consumer])

        generator1.outputs["Value"] >> merger.inputs["In1"]
        generator1.outputs["Weight"] >> merger.inputs["In1_weight"]

        generator2.outputs["Value"] >> merger.inputs["In2"]
        generator2.outputs["Weight"] >> merger.inputs["In2_weight"]

        merger.outputs["WeightedSum"] >> consumer.inputs["WeightedSum"]

        composition.run(start_time=start, end_time=start + timedelta(days=30))

        self.assertEqual(
            fm.data.get_units(consumer.data["WeightedSum"]), fm.UNITS("mm")
        )

    def test_weighted_sum_fail_grid(self):
        start = datetime(2000, 1, 1)
        source_grid = fm.UniformGrid((20, 15))
        source_grid_2 = fm.UniformGrid((20, 10))

        generator1 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=1),
        )
        generator2 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid_2),
                    fm.Info(None, grid=source_grid_2, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid_2),
                    fm.Info(None, source_grid_2),
                ),
            },
            start=start,
            step=timedelta(days=5),
        )

        merger = WeightedSum(inputs=["In1", "In2"])

        consumer = DebugConsumer(
            inputs={"WeightedSum": fm.Info(None, grid=None, units=None)},
            start=start,
            step=timedelta(days=1),
        )

        composition = fm.Composition([generator1, generator2, merger, consumer])

        generator1.outputs["Value"] >> merger.inputs["In1"]
        generator1.outputs["Weight"] >> merger.inputs["In1_weight"]

        generator2.outputs["Value"] >> merger.inputs["In2"]
        generator2.outputs["Weight"] >> merger.inputs["In2_weight"]

        merger.outputs["WeightedSum"] >> consumer.inputs["WeightedSum"]

        with self.assertRaises(finam.errors.FinamMetaDataError):
            composition.run(start_time=start, end_time=start + timedelta(days=30))

    def test_weighted_sum_fail_grid_weights(self):
        start = datetime(2000, 1, 1)
        source_grid = fm.UniformGrid((20, 15))
        source_grid_2 = fm.UniformGrid((20, 10))

        generator1 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=1),
        )
        generator2 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid_2),
                    fm.Info(None, source_grid_2),
                ),
            },
            start=start,
            step=timedelta(days=5),
        )

        merger = WeightedSum(inputs=["In1", "In2"])

        consumer = DebugConsumer(
            inputs={"WeightedSum": fm.Info(None, grid=None, units=None)},
            start=start,
            step=timedelta(days=1),
        )

        composition = fm.Composition([generator1, generator2, merger, consumer])

        generator1.outputs["Value"] >> merger.inputs["In1"]
        generator1.outputs["Weight"] >> merger.inputs["In1_weight"]

        generator2.outputs["Value"] >> merger.inputs["In2"]
        generator2.outputs["Weight"] >> merger.inputs["In2_weight"]

        merger.outputs["WeightedSum"] >> consumer.inputs["WeightedSum"]

        with self.assertRaises(finam.errors.FinamMetaDataError):
            composition.run(start_time=start, end_time=start + timedelta(days=30))

    def test_weighted_sum_fail_units(self):
        start = datetime(2000, 1, 1)
        source_grid = fm.UniformGrid((20, 15))

        generator1 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=1),
        )
        generator2 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="s"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=5),
        )

        merger = WeightedSum(inputs=["In1", "In2"])

        consumer = DebugConsumer(
            inputs={"WeightedSum": fm.Info(None, grid=None, units=None)},
            start=start,
            step=timedelta(days=1),
        )

        composition = fm.Composition([generator1, generator2, merger, consumer])

        generator1.outputs["Value"] >> merger.inputs["In1"]
        generator1.outputs["Weight"] >> merger.inputs["In1_weight"]

        generator2.outputs["Value"] >> merger.inputs["In2"]
        generator2.outputs["Weight"] >> merger.inputs["In2_weight"]

        merger.outputs["WeightedSum"] >> consumer.inputs["WeightedSum"]

        with self.assertRaises(finam.errors.FinamMetaDataError):
            composition.run(start_time=start, end_time=start + timedelta(days=30))

    def test_weighted_sum_fail_units_weights(self):
        start = datetime(2000, 1, 1)
        source_grid = fm.UniformGrid((20, 15))

        generator1 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid),
                ),
            },
            start=start,
            step=timedelta(days=1),
        )
        generator2 = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, grid=source_grid, units="mm"),
                ),
                "Weight": (
                    lambda t: generate_grid(source_grid),
                    fm.Info(None, source_grid, units="mm"),
                ),
            },
            start=start,
            step=timedelta(days=5),
        )

        merger = WeightedSum(inputs=["In1", "In2"])

        consumer = DebugConsumer(
            inputs={"WeightedSum": fm.Info(None, grid=None, units=None)},
            start=start,
            step=timedelta(days=1),
        )

        composition = fm.Composition([generator1, generator2, merger, consumer])

        generator1.outputs["Value"] >> merger.inputs["In1"]
        generator1.outputs["Weight"] >> merger.inputs["In1_weight"]

        generator2.outputs["Value"] >> merger.inputs["In2"]
        generator2.outputs["Weight"] >> merger.inputs["In2_weight"]

        merger.outputs["WeightedSum"] >> consumer.inputs["WeightedSum"]

        with self.assertRaises(finam.errors.FinamMetaDataError):
            composition.run(start_time=start, end_time=start + timedelta(days=30))


if __name__ == "__main__":
    unittest.main()



"""
Adapters for statistics over grids.
"""
import numpy as np

from ..data import get_magnitude
from ..data.grid_spec import UniformGrid
from ..sdk import Adapter

__all__ = [
    "Histogram",
]


class Histogram(Adapter):
    """Calculates a histogram over grid values.

    Results are returned as a 1-D uniform grid.

    Parameters
    ----------
    lower : float
        Lower bound of the histogram.
    upper : float
        Upper bound of the histogram.
    bins : int, optional
        Number of bins. Default: 10.
    density : bool, optional
        Calculates densities instead of counts. Default: False.

    Returns
    -------

    UniformGrid
        A 1-D grid using :class:`.UniformGrid` as specification.
    """

    def __init__(self, lower, upper, bins=10, density=False):
        super().__init__()
        self.lower = lower
        self.upper = upper
        self.bins = bins
        self.density = density

        step = (upper - lower) / bins
        self.grid = UniformGrid(dims=(bins + 1,), spacing=(step,), origin=(lower,))

    def _get_data(self, time, target):
        d = get_magnitude(self.pull_data(time, target))

        values = np.histogram(
            d, bins=self.bins, range=(self.lower, self.upper), density=self.density
        )
        return values[0]

    def _get_info(self, info):
        info = info.copy_with(grid=None)
        in_info = self.exchange_info(info)
        out_info = in_info.copy_with(grid=self.grid, units="")
        return out_info



"""
Unit tests for regridding with masked data.
"""

import unittest
from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

from finam import Composition, FinamDataError, Info, Mask, UniformGrid, UnstructuredGrid
from finam import data as fdata
from finam.adapters.regrid import RegridLinear, RegridNearest
from finam.components import StaticSimplexNoise, debug


def get_mask(points, rad=1.5):
    return (points[:, 0] - 2.5) ** 2 + (points[:, 1] - 2) ** 2 > rad**2


class TestRegridMask(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        here = Path(__file__).parent
        points = np.loadtxt(here / "square_5x4" / "points.txt", dtype=float)
        cells = np.loadtxt(here / "square_5x4" / "cells.txt", dtype=int)
        types = np.loadtxt(here / "square_5x4" / "types.txt", dtype=int)

        self.in_grid = UnstructuredGrid(points, cells, types)
        self.out_grid = UniformGrid((25, 20), spacing=(0.2, 0.2))

        self.in_mask = get_mask(self.in_grid.cell_centers, rad=1.5)
        omask = get_mask(self.out_grid.cell_centers, rad=2.5)
        self.out_mask = fdata.from_compressed(
            omask, self.out_grid.data_shape, self.out_grid.order
        )

    def test_regrid_nearest_out_mask(self):
        in_info = Info(grid=self.in_grid, units="", mask=self.in_mask)
        source = StaticSimplexNoise(in_info, 0.15, 3, 0.5)
        sink = debug.DebugPushConsumer({"Input": Info()})
        regrid = RegridNearest(out_grid=self.out_grid, out_mask=self.out_mask)
        composition = Composition([source, sink])

        (source.outputs["Noise"] >> regrid >> sink.inputs["Input"])
        composition.connect()

        info = sink.inputs["Input"].info
        data = sink.data["Input"][0, ...]
        assert_array_equal(info.mask, data.mask)
        assert_array_equal(info.mask, self.out_mask)

        i_data = source.outputs["Noise"].data[0][1].magnitude.compressed()
        o_data = data.magnitude.compressed()
        self.assertAlmostEqual(i_data.mean(), o_data.mean(), 2)

    def test_regrid_nearest_filled(self):
        in_info = Info(grid=self.in_grid, units="", mask=self.in_mask)
        source = StaticSimplexNoise(in_info, 0.15, 3, 0.5)
        sink = debug.DebugPushConsumer({"Input": Info()})
        regrid = RegridNearest(out_grid=self.out_grid, out_mask=Mask.NONE)
        composition = Composition([source, sink])

        (source.outputs["Noise"] >> regrid >> sink.inputs["Input"])
        composition.connect()

        info = sink.inputs["Input"].info
        data = sink.data["Input"][0, ...].magnitude
        self.assertEqual(info.mask, Mask.NONE)
        self.assertFalse(fdata.is_masked_array(data))

        i_data = source.outputs["Noise"].data[0][1].magnitude.compressed()
        self.assertAlmostEqual(i_data.mean(), data.mean(), 1)

    def test_regrid_linear_determine_mask(self):
        in_info = Info(grid=self.in_grid, units="", mask=self.in_mask)
        source = StaticSimplexNoise(in_info, 0.15, 3, 0.5)
        sink = debug.DebugPushConsumer({"Input": Info()})
        regrid = RegridLinear(
            out_grid=self.out_grid, fill_with_nearest=False, out_mask=None
        )
        composition = Composition([source, sink])

        (source.outputs["Noise"] >> regrid >> sink.inputs["Input"])
        composition.connect()

        info = sink.inputs["Input"].info
        data = sink.data["Input"][0, ...]
        assert_array_equal(info.mask, data.mask)
        self.assertEqual(np.sum(data.mask), 306)

        i_data = source.outputs["Noise"].data[0][1].magnitude.compressed()
        o_data = data.magnitude.compressed()
        self.assertAlmostEqual(i_data.mean(), o_data.mean(), 2)

    def test_regrid_linear_error_domain(self):
        in_info = Info(grid=self.in_grid, units="", mask=self.in_mask)
        source = StaticSimplexNoise(in_info, 0.15, 3, 0.5)
        sink = debug.DebugPushConsumer({"Input": Info()})
        regrid = RegridLinear(
            out_grid=self.out_grid, fill_with_nearest=False, out_mask=Mask.NONE
        )
        composition = Composition([source, sink])

        (source.outputs["Noise"] >> regrid >> sink.inputs["Input"])

        # not covering domain without fill
        with self.assertRaises(FinamDataError):
            composition.connect()

    def test_regrid_linear_filled(self):
        in_info = Info(grid=self.in_grid, units="", mask=self.in_mask)
        source = StaticSimplexNoise(in_info, 0.15, 3, 0.5)
        sink = debug.DebugPushConsumer({"Input": Info()})
        regrid = RegridLinear(
            out_grid=self.out_grid, fill_with_nearest=True, out_mask=Mask.NONE
        )
        composition = Composition([source, sink])

        (source.outputs["Noise"] >> regrid >> sink.inputs["Input"])
        composition.connect()

        info = sink.inputs["Input"].info
        data = sink.data["Input"][0, ...].magnitude
        self.assertEqual(info.mask, Mask.NONE)
        self.assertFalse(fdata.is_masked_array(data))

        i_data = source.outputs["Noise"].data[0][1].magnitude.compressed()
        self.assertAlmostEqual(i_data.mean(), data.mean(), 1)

    def test_regrid_linear_filled_mask(self):
        in_info = Info(grid=self.in_grid, units="", mask=self.in_mask)
        source = StaticSimplexNoise(in_info, 0.15, 3, 0.5)
        sink = debug.DebugPushConsumer({"Input": Info()})
        regrid = RegridLinear(
            out_grid=self.out_grid, fill_with_nearest=True, out_mask=self.out_mask
        )
        composition = Composition([source, sink])

        (source.outputs["Noise"] >> regrid >> sink.inputs["Input"])
        composition.connect()

        info = sink.inputs["Input"].info
        data = sink.data["Input"][0, ...]
        assert_array_equal(info.mask, data.mask)
        assert_array_equal(info.mask, self.out_mask)

        i_data = source.outputs["Noise"].data[0][1].magnitude.compressed()
        o_data = data.magnitude.compressed()
        self.assertAlmostEqual(i_data.mean(), o_data.mean(), 2)


if __name__ == "__main__":
    unittest.main()
